#
# Qt gui for viewing DataBlock2d data files.
#
# This file is part of Myokit.
# See http://myokit.org for copyright, sharing, and licensing details.
#
import configparser
import gc
import os
import sys
import traceback

from myokit.gui import QtWidgets, QtGui, QtCore, Qt

import myokit
import myokit.gui
import myokit.gui.progress
import myokit.formats.axon
import myokit.formats.wcp

# Matplotlib (must be imported _after_ gui has had chance to set backend)
import matplotlib
import matplotlib.figure
from myokit.gui import matplotlib_backend as backend

# NumPy
import numpy as np

# SciPy: Only used to load matlab files.
try:
    import scipy.io
    has_scipy = True
except ImportError:
    has_scipy = False


# Application title
TITLE = 'Myokit DataLog Viewer (PROTOTYPE)'

# Application icon
# def icon():
#    icons = [
#        'icon-datalog-viewer.ico',
#        'icon-datalog-viewer-16.xpm',
#        'icon-datalog-viewer-24.xpm',
#        'icon-datalog-viewer-32.xpm',
#        'icon-datalog-viewer-48.xpm',
#        'icon-datalog-viewer-64.xpm',
#        'icon-datalog-viewer-96.xpm',
#        'icon-datalog-viewer-128.xpm',
#        'icon-datalog-viewer-256.xpm',
#        ]
#    icon = QtGui.QIcon()
#    for i in icons:
#        icon.addFile(os.path.join(myokit.DIR_DATA, 'gui', i))
#    return icon

# Settings file
SETTINGS_FILE = os.path.join(myokit.DIR_USER, 'DataLogViewer.ini')
# Number of recent files to display
# N_RECENT_FILES = 5

# About
ABOUT = '<h1>' + TITLE + '</h1>' + """
<p>
    The DataLog viewer is a PROTOTYPE utility to examine time series data.
    At the moment, exclusively WinWCP, ABF and CSV files.
</p>
<p>
    System info:
    <br />Python: PYTHON
    <br />Using the BACKEND GUI backend.
</p>
""".replace('BACKEND', myokit.gui.backend).replace('PYTHON', sys.version)

# License
LICENSE = myokit.LICENSE_HTML

# File filters
FILTER_ABF = 'ABF files (*.abf *.pro)'
FILTER_ATF = 'ATF files (*.atf)'
FILTER_CSV = 'CSV files (*.csv)'
FILTER_MAT = 'Matlab files (*.mat)'
FILTER_DAT = 'PatchMaster files (*.dat)'
FILTER_TXT = 'Text files (*.txt)'
FILTER_WCP = 'WCP files (*.wcp)'
FILTER_ZIP = 'Zipped DataLog files (*.zip)'
FILTER_ANY = 'All files (*.*)'
FILTER_ALL = 'Data files (*.abf *.csv *.dat *.mat *.pro *.txt *.wcp *.zip)'
FILTER_LIST = ';;'.join([
    FILTER_ALL,
    FILTER_ABF,
    FILTER_CSV,
    FILTER_DAT,
    FILTER_MAT,
    FILTER_TXT,
    FILTER_WCP,
    FILTER_ZIP,
    FILTER_ANY,
])


class DataLogViewer(myokit.gui.MyokitApplication):
    """
    Graphical interface for viewing DataLog data.
    """
    def __init__(self, *filenames):
        super().__init__()

        # Set Title, icon
        self.setWindowTitle(TITLE + ' ' + myokit.__version__)

        # Set size, center
        self.resize(800, 600)
        qr = self.frameGeometry()
        cp = QtGui.QGuiApplication.primaryScreen().availableGeometry().center()
        qr.moveCenter(cp)
        self.move(qr.topLeft())

        # Add widget for file tabs
        self._tabs = TabWidget()
        self._tabs.setTabsClosable(True)
        self._tabs.tabCloseRequested.connect(self.action_close)
        self._tabs.currentChanged.connect(self.fileTabChangeEvent)
        self.setCentralWidget(self._tabs)

        # Menu bar
        self.create_menu()

        # Tool bar
        self.create_toolbar()

        # Status bar
        self.statusBar().showMessage('Ready')

        # Current path
        self._path = QtCore.QDir.currentPath()

        # Load settings from ini file
        self.load_config()

        # File-loading methods
        self._load_actions = {
            '.abf': self.load_abf_file,
            '.atf': self.load_atf_file,
            '.csv': self.load_datalog,
            '.dat': self.load_dat_file,
            '.pro': self.load_abf_file,
            '.txt': self.load_txt_file,
            '.wcp': self.load_wcp_file,
            '.zip': self.load_datalog,
        }
        if has_scipy:
            self._load_actions['.mat'] = self.load_mat_file

        # Load any selected files
        for filename in filenames:
            self.load_file(filename)
        tc = self._tabs.count()
        if tc > 0:
            if tc > 1:
                self._tool_next_file.setEnabled(True)
                self._tool_prev_file.setEnabled(True)
            self._tabs.setCurrentIndex(0)

    def action_about(self):
        """
        Displays the about dialog.
        """
        QtWidgets.QMessageBox.about(self, TITLE, ABOUT)

    def action_close(self, index):
        """
        Called when a tab should be closed
        """
        # Remove tab
        tab = self._tabs.widget(index)
        self._tabs.removeTab(index)

        # Update buttons
        if self._tabs.count() < 2:
            self._tool_next_file.setEnabled(False)
            self._tool_prev_file.setEnabled(False)

        # Delete tab
        if tab is not None:
            tab.deleteLater()
        gc.collect()
        del tab

    def action_first_var(self):
        """ Select the first variable in the current file. """
        tab = self._tabs.currentWidget()
        if tab:
            tab.first()

    def action_last_var(self):
        """ Select the last variable in the current file. """
        tab = self._tabs.currentWidget()
        if tab:
            tab.last()

    def action_license(self):
        """
        Displays this program's licensing information.
        """
        QtWidgets.QMessageBox.about(self, TITLE, LICENSE)

    def action_open(self):
        """
        Let the user select and open a file.
        """
        filenames = QtWidgets.QFileDialog.getOpenFileNames(
            self, 'Open data file', self._path, filter=FILTER_LIST)[0]
        if filenames:
            # Save current number of tabs
            tab_count = self._tabs.count()

            # Load files
            for filename in filenames:
                self.load_file(str(filename))

            # If loading went ok, show first of newly loaded files
            tab_count_new = self._tabs.count()
            if tab_count_new > tab_count:
                self._tabs.setCurrentIndex(tab_count)

                # Enable next/previous file menu items
                if tab_count_new > 1:
                    self._tool_next_file.setEnabled(True)
                    self._tool_prev_file.setEnabled(True)

    def action_next_file(self):
        """
        Select the next open file.
        """
        self._tabs.next()

    def action_next_var(self):
        """
        Select the next variable in the selected file.
        """
        tab = self._tabs.currentWidget()
        if tab:
            tab.next()

    def action_prev_file(self):
        """
        Select the previous open file.
        """
        self._tabs.previous()

    def action_prev_var(self):
        """
        Select the previous variable in the selected file
        """
        tab = self._tabs.currentWidget()
        if tab:
            tab.previous()

    def closeEvent(self, event=None):
        """
        Called when window is closed. To force a close (and trigger this
        function, call self.close())
        """
        # Save configuration
        self.save_config()
        if event:
            # Accept the event, close the window
            event.accept()
            # Ignore the event, window stays open
            #event.ignore()

    def create_menu(self):
        """
        Creates this widget's menu.
        """
        self._menu = self.menuBar()
        # File menu
        self._menu_file = self._menu.addMenu('&File')
        # File > Open
        self._tool_open = QtGui.QAction('&Open', self)
        self._tool_open.setShortcut('Ctrl+O')
        self._tool_open.setStatusTip('Open a file')
        self._tool_open.setIcon(QtGui.QIcon.fromTheme('document-open'))
        self._tool_open.triggered.connect(self.action_open)
        self._menu_file.addAction(self._tool_open)
        # File > ----
        self._menu_file.addSeparator()
        # File > Quit
        self._tool_exit = QtGui.QAction('&Quit', self)
        self._tool_exit.setShortcut('Ctrl+Q')
        self._tool_exit.setStatusTip('Exit application.')
        self._tool_exit.setIcon(QtGui.QIcon.fromTheme('application-exit'))
        self._tool_exit.triggered.connect(self.close)
        self._menu_file.addAction(self._tool_exit)
        # View menu
        self._menu_view = self._menu.addMenu('&View')
        # View > Next file
        self._tool_next_file = QtGui.QAction('Next file', self)
        self._tool_next_file.setShortcut('Ctrl+PgDown')
        self._tool_next_file.setStatusTip('Select the next open file')
        self._tool_next_file.triggered.connect(self.action_next_file)
        self._tool_next_file.setEnabled(False)
        self._menu_view.addAction(self._tool_next_file)
        # View > Previous file
        self._menu_view.addAction(self._tool_next_file)
        self._tool_prev_file = QtGui.QAction('Previous file', self)
        self._tool_prev_file.setShortcut('Ctrl+PgUp')
        self._tool_prev_file.setStatusTip('Select the previous open file')
        self._tool_prev_file.triggered.connect(self.action_prev_file)
        self._tool_prev_file.setEnabled(False)
        self._menu_view.addAction(self._tool_prev_file)
        # View > ----
        self._menu_view.addSeparator()
        # View > Next variable
        self._tool_next_var = QtGui.QAction('Next variable', self)
        self._tool_next_var.setShortcut('PgDown')
        self._tool_next_var.setStatusTip('Show the next variable')
        self._tool_next_var.triggered.connect(self.action_next_var)
        self._tool_next_var.setEnabled(False)
        self._menu_view.addAction(self._tool_next_var)
        # View > Previous var
        self._tool_prev_var = QtGui.QAction('Previous variable', self)
        self._tool_prev_var.setShortcut('PgUp')
        self._tool_prev_var.setStatusTip('Show the previous variable')
        self._tool_prev_var.triggered.connect(self.action_prev_var)
        self._tool_prev_var.setEnabled(False)
        self._menu_view.addAction(self._tool_prev_var)
        # View > First variable
        self._tool_first_var = QtGui.QAction('First variable', self)
        self._tool_first_var.setShortcut('Home')
        self._tool_first_var.setStatusTip('Show the first variable')
        self._tool_first_var.triggered.connect(self.action_first_var)
        self._tool_first_var.setEnabled(False)
        self._menu_view.addAction(self._tool_first_var)
        # View > Last var
        self._tool_last_var = QtGui.QAction('Last variable', self)
        self._tool_last_var.setShortcut('End')
        self._tool_last_var.setStatusTip('Show the last variable')
        self._tool_last_var.triggered.connect(self.action_last_var)
        self._tool_last_var.setEnabled(False)
        self._menu_view.addAction(self._tool_last_var)
        # Help menu
        self._menu_help = self._menu.addMenu('&Help')
        # Help > About
        self._tool_about = QtGui.QAction('&About', self)
        self._tool_about.setStatusTip('View information about this program.')
        self._tool_about.triggered.connect(self.action_about)
        self._menu_help.addAction(self._tool_about)
        # Help > License
        self._tool_license = QtGui.QAction('&License', self)
        self._tool_license.setStatusTip('View this program\'s license info.')
        self._tool_license.triggered.connect(self.action_license)
        self._menu_help.addAction(self._tool_license)

    def create_toolbar(self):
        """
        Creates this widget's toolbar
        """
        self._toolbar = self.addToolBar('tools')
        self._toolbar.setFloatable(False)
        self._toolbar.setMovable(False)
        self._toolbar.setToolButtonStyle(
            Qt.ToolButtonStyle.ToolButtonTextBesideIcon)
        self._toolbar.addAction(self._tool_open)
        #self._toolbar.addSeparator()

    def load_config(self):
        """
        Loads the user configuration from an ini file.
        """
        # Read ini file
        config = configparser.RawConfigParser()
        try:
            config.read(os.path.expanduser(SETTINGS_FILE))
        except configparser.ParsingError:
            # Partially read config causes all sorts of errors, so discard
            config = configparser.RawConfigParser()

        # Window dimensions and location
        if config.has_section('window'):
            g = self.geometry()

            def getor(name, alt):
                if config.has_option('window', name):
                    return int(config.get('window', name))
                return alt

            x = getor('x', g.x())
            y = getor('y', g.y())
            w = getor('w', g.width())
            h = getor('h', g.height())
            self.setGeometry(x, y, w, h)

        # Current files, directory, etc
        if config.has_section('files'):
            if config.has_option('files', 'path'):
                path = config.get('files', 'path')
                if os.path.isdir(path):
                    self._path = path

    def load_file(self, filename):
        """
        Loads a data file.
        """
        root, ext = os.path.splitext(os.path.basename(filename))
        try:
            action = self._load_actions[ext.lower()]
        except KeyError:
            QtWidgets.QMessageBox.critical(
                self, TITLE, 'File format not recognized: ' + ext)
            return
        action(filename)

    def load_abf_file(self, filename):
        """ Loads an ABF file. """
        try:
            abf = myokit.formats.axon.AbfFile(filename)
        except Exception:
            e = traceback.format_exc()
            QtWidgets.QMessageBox.critical(self, TITLE, e)
            return
        self._path = os.path.dirname(filename)
        self._tabs.addTab(AbfTab(self, abf), os.path.basename(filename))

    def load_atf_file(self, filename):
        """ Loads an ATF file. """
        try:
            atf = myokit.formats.axon.AtfFile(filename)
        except Exception:
            e = traceback.format_exc()
            QtWidgets.QMessageBox.critical(self, TITLE, e)
            return
        self._path = os.path.dirname(filename)
        self._tabs.addTab(AtfTab(self, atf), os.path.basename(filename))

    def load_dat_file(self, filename):
        """ Loads a PatchMaster dat file. """

        pbar = myokit.gui.progress.ProgressBar(
            self, 'Loading groups and series')
        pbar.show()
        reporter = pbar.reporter()
        reporter.enter()

        flag = QtCore.QEventLoop.ProcessEventsFlag.ExcludeUserInputEvents
        QtWidgets.QApplication.processEvents(flag)

        try:
            with myokit.formats.heka.PatchMasterFile(filename) as f:
                n = sum([len(list(g.complete_series())) for g in f])
                i = 0
                stop = False
                for group in f:
                    for series in group.complete_series():
                        self._tabs.addTab(
                            PatchMasterTab(self, series),
                            f'{group.label()} {series.label()}')
                        i += 1
                        if not reporter.update(i / n):
                            stop = True
                            break
                        QtWidgets.QApplication.processEvents(flag)
                    if stop:
                        break
                self._path = os.path.dirname(filename)
        except Exception:
            e = traceback.format_exc()
            QtWidgets.QMessageBox.critical(self, TITLE, e)
            return
        finally:
            reporter.exit()
            pbar.close()
            pbar.deleteLater()

    def load_datalog(self, filename):
        """ Loads a DataLog from csv or zip file. """
        try:
            if filename[-4:].lower() == '.csv':
                log = myokit.DataLog.load_csv(filename)
            else:
                log = myokit.DataLog.load(filename)
            if log.time_key is None:
                raise Exception('Log must contain a suitable time variable.')
        except Exception:
            e = traceback.format_exc()
            QtWidgets.QMessageBox.critical(self, TITLE, e)
            return
        self._path = os.path.dirname(filename)
        name = os.path.basename(filename)
        self._tabs.addTab(CsvTab(self, log, name), name)

    def load_mat_file(self, filename):
        """
        Loads a Matlab file.

        This method requires ``SciPy`` to be installed.
        """
        try:
            mat = scipy.io.loadmat(filename)
        except Exception:
            e = traceback.format_exc()
            QtWidgets.QMessageBox.critical(self, TITLE, e)
            return
        self._path = os.path.dirname(filename)
        name = os.path.basename(filename)
        self._tabs.addTab(MatTab(self, mat, name), name)

    def load_txt_file(self, filename):
        """ Loads a text file. """
        try:
            data = np.loadtxt(filename)
        except Exception:
            e = traceback.format_exc()
            QtWidgets.QMessageBox.critical(self, TITLE, e)
            return
        self._path = os.path.dirname(filename)
        name = os.path.basename(filename)
        self._tabs.addTab(TxtTab(self, data, name), name)

    def load_wcp_file(self, filename):
        """ Loads a WinWCP file. """
        try:
            wcp = myokit.formats.wcp.WcpFile(filename)
        except Exception:
            e = traceback.format_exc()
            QtWidgets.QMessageBox.critical(self, TITLE, e)
            return
        self._path = os.path.dirname(filename)
        self._tabs.addTab(WcpTab(self, wcp), os.path.basename(filename))

    def save_config(self):
        """
        Saves the user configuration to an ini file.
        """
        config = configparser.RawConfigParser()

        # Window dimensions and location
        config.add_section('window')
        g = self.geometry()
        config.set('window', 'x', str(g.x()))
        config.set('window', 'y', str(g.y()))
        config.set('window', 'w', str(g.width()))
        config.set('window', 'h', str(g.height()))

        # Current files, directory, etc
        config.add_section('files')
        config.set('files', 'path', self._path)

        # Write configuration to ini file
        inifile = os.path.expanduser(SETTINGS_FILE)
        with open(inifile, 'w') as configfile:
            config.write(configfile)

    def show(self):
        """
        Shows this viewer.
        """
        super().show()
        QtWidgets.QApplication.processEvents()

    def fileTabChangeEvent(self, index):
        """
        Different file tab selected.
        """
        if index >= 0:
            tab = self._tabs.widget(index)
            if tab.count() > 1:
                self._tool_first_var.setEnabled(True)
                self._tool_last_var.setEnabled(True)
                self._tool_next_var.setEnabled(True)
                self._tool_prev_var.setEnabled(True)
                return
        self._tool_first_var.setEnabled(False)
        self._tool_last_var.setEnabled(False)
        self._tool_prev_var.setEnabled(False)
        self._tool_next_var.setEnabled(False)


class TabWidget(QtWidgets.QTabWidget):
    """ Generic tab widget with first/last next/previous methods. """

    def first(self):
        """ Select the first widget. """
        self.setCurrentIndex(0)

    def last(self):
        """ Select the last widget. """
        self.setCurrentIndex(self.count() - 1)

    def next(self):
        """ Select the next widget. """
        n = self.count()
        if n < 2:
            return
        i = self.currentIndex() + 1
        self.setCurrentIndex(0 if i >= n else i)

    def previous(self):
        """ Select the previous widget. """
        n = self.count()
        if n < 2:
            return
        i = self.currentIndex() - 1
        self.setCurrentIndex(n - 1 if i < 0 else i)


class GraphTabWidget(TabWidget):
    """ Tab widget to graph a data source. """

    def __init__(self, parent):
        super().__init__(parent)

        self.setTabsClosable(False)
        self.setTabPosition(self.TabPosition.East)

        self._figures = []
        self._axes = []

    def deleteLater(self):
        """ Deletes this tab (later). """
        for figure in self._figures:
            figure.clear()
        for axes in self._axes:
            axes.cla()
        del self._figures, self._axes
        gc.collect()
        super().deleteLater()


class SweepSourceTab(GraphTabWidget):
    """ A tab widget for sources implementing the SweepSource interface. """

    def __init__(self, parent, source):
        super().__init__(parent)

        # Add A/D
        for i in range(source.channel_count()):
            self._add_graph_tab(source, i)

        # Add D/A
        for i in range(source.da_count()):
            self._add_graph_tab(source, i, True)

        # Add meta data
        self._add_meta_tab(source)

    def _add_graph_tab(self, source, index, da=False):
        """ Adds a tab for a graph. """

        # Create widget
        widget = QtWidgets.QWidget(self)

        # Create figure
        figure = matplotlib.figure.Figure()
        canvas = backend.FigureCanvasQTAgg(figure)
        canvas.setParent(widget)
        toolbar = backend.NavigationToolbar2QT(canvas, widget)

        # Draw signal
        join_sweeps = not source.equal_length_sweeps()
        if da:
            name = source.da_names(index)
            units = source.da_units(index)
            times, values = source.da(index, join_sweeps)
        else:
            name = source.channel_names(index)
            units = source.channel_units(index)
            times, values = source.channel(index, join_sweeps)

        axes = figure.add_subplot(1, 1, 1)
        axes.set_xlabel(f'Time {source.time_unit()}')
        axes.set_ylabel(f'{name} {units}')
        if join_sweeps:
            axes.plot(times, values)
        else:
            for v in values:
                axes.plot(times[0], v)

        # Store for later deletion
        self._figures.append(figure)
        self._axes.append(axes)

        # Create a layout
        vbox = QtWidgets.QVBoxLayout()
        vbox.addWidget(canvas)
        vbox.addWidget(toolbar)
        widget.setLayout(vbox)

        # Add tab
        self.addTab(widget, name)

    '''
    def debug_tab(self, channel_index, da=True):
        """ Add a tab graphing data using the DataLog method. """
        widget = QtWidgets.QWidget(self)

        # Create widget
        widget = QtWidgets.QWidget(self)

        # Create figure
        figure = matplotlib.figure.Figure()
        figure.suptitle(self._abf.filename())
        canvas = backend.FigureCanvasQTAgg(figure)
        canvas.setParent(widget)
        toolbar = backend.NavigationToolbar2QT(canvas, widget)

        # Draw lines
        p = self._abf.da_protocol(channel_index, tu='s', vu='V')
        times, _ = self._abf.da(channel_index)
        for t in times:
            d = p.log_for_interval(t[0], t[-1], for_drawing=True).npview()
            axes.plot(d['time'] - t[0], d['pace'])

        # Create a layout
        vbox = QtWidgets.QVBoxLayout()
        vbox.addWidget(canvas)
        vbox.addWidget(toolbar)
        widget.setLayout(vbox)

        # Add tab
        self.addTab(widget, name)
'''

    def _add_meta_tab(self, source):

        meta = source.meta_str(True)
        if meta:
            widget = QtWidgets.QTextEdit(self)
            widget.setText(meta)
            widget.setReadOnly(True)
            self.addTab(widget, 'info')


class AbfTab(SweepSourceTab):
    """ A widget displaying an ABF file. """
    pass


class AtfTab(GraphTabWidget):
    """ A widget displaying an ATF file. """
    def __init__(self, parent, atf):
        super().__init__(parent)
        self._atf = atf

        keys = list(self._atf.keys())
        if len(keys) > 1:
            time = keys[0]  # Time is always first (and regularly sampled)
            for key in keys[1:]:
                self.addTab(self.create_graph_tab(time, key), key)
        self.addTab(self.create_info_tab(), 'Info')
        del self._atf

    def create_graph_tab(self, time, key):
        """
        Creates a widget displaying a graph.
        """
        widget = QtWidgets.QWidget(self)

        # Create figure
        figure = matplotlib.figure.Figure()
        figure.suptitle(self._atf.filename())
        canvas = backend.FigureCanvasQTAgg(figure)
        canvas.setParent(widget)
        axes = figure.add_subplot(1, 1, 1)
        toolbar = backend.NavigationToolbar2QT(canvas, widget)

        # Draw lines
        axes.plot(self._atf[time], self._atf[key])

        # Create a layout
        vbox = QtWidgets.QVBoxLayout()
        vbox.addWidget(canvas)
        vbox.addWidget(toolbar)
        widget.setLayout(vbox)
        self._figures.append(figure)
        self._axes.append(axes)

        # Return widget
        return widget

    def create_info_tab(self):
        """ Creates a tab displaying information about the file. """
        widget = QtWidgets.QTextEdit(self)
        widget.setText(self._atf.info())
        widget.setReadOnly(True)
        return widget


class CsvTab(GraphTabWidget):
    """
    A widget displaying a CSV file.

    The given log must have a time variable set.
    """
    def __init__(self, parent, log, filename):
        super().__init__(parent)

        self._log = log.npview()
        self._filename = filename

        # Check time key was found
        time = log.time_key()
        try:
            self._time = log.time()
        except myokit.InvalidDataLogError:
            if time is None:
                QtWidgets.QMessageBox.critical(
                    self, TITLE,
                    'Unable to load file: no time key set in this log.')
                return
            else:
                raise

        # Check that time series aren't empty
        if log.length() == 0:
            QtWidgets.QMessageBox.critical(
                self, TITLE, 'Unable to load file: no data found.')
            return

        # Overlapping sweeps or neighboring cells?
        keys = []
        groups = {}
        for key in log.keys():
            if key == time:
                continue
            index, var = myokit.split_key(key)
            if index:
                group = groups.get(var, None)
                if group is None:
                    groups[var] = [index]
                    keys.append(var)
                else:
                    group.append(index)
            else:
                keys.append(var)

        # Add tab for each column
        for k in keys:
            self.addTab(self.create_graph_tab(k, groups.get(k)), k)

    def create_graph_tab(self, key, indices=None):
        """ Creates a widget displaying the data stored under ``key``. """
        widget = QtWidgets.QWidget(self)

        # Create figure
        figure = matplotlib.figure.Figure()
        figure.suptitle(self._filename)
        canvas = backend.FigureCanvasQTAgg(figure)
        canvas.setParent(widget)
        axes = figure.add_subplot(1, 1, 1)
        axes.set_title(key)
        toolbar = backend.NavigationToolbar2QT(canvas, widget)

        # Draw lines
        if indices is None:
            axes.plot(self._time, self._log[key])
        else:
            for i in indices:
                axes.plot(self._time, self._log[i + key])

        # Create a layout
        vbox = QtWidgets.QVBoxLayout()
        vbox.addWidget(canvas)
        vbox.addWidget(toolbar)
        widget.setLayout(vbox)
        self._figures.append(figure)
        self._axes.append(axes)
        return widget


class MatTab(GraphTabWidget):
    """ A widget displaying a .mat file. """
    def __init__(self, parent, mat, filename):
        super().__init__(parent)

        self._filename = filename

        # Find usable data
        for key in mat.keys():
            if key[:1] == '_':
                continue
            time, data = None, None
            data = mat[key]
            if np.prod(data.shape) == np.max(data.shape):
                # 1d data
                time = None
                data = data.reshape((np.max(data.shape),))
            elif len(data.shape) == 2 and (
                    np.prod(data.shape) == 2 * np.max(data.shape)):
                # 2d data: Only allow len(shape) == 2, otherwise too many cases
                if data.shape[0] == 2:
                    time = data[0]
                    data = data[1]
                else:
                    time = data[:, 0]
                    data = data[:, 1]
                # Check time is increasing
                if np.any(time[1:] < time[:-1]):
                    time, data = data, time
                if np.any(time[1:] < time[:-1]):
                    time, data = None, None
            if data is None:
                continue
            # Create tab
            tab = self.create_graph_tab(time, data)
            self.addTab(tab, key)

        # Nothing that can be used? Show error
        if len(self._figures) == 0:
            QtWidgets.QMessageBox.critical(
                self, TITLE,
                'Unable to load file: no usable data found.')
            return

    def create_graph_tab(self, time, data):
        """
        Creates a widget displaying a time series.
        """
        widget = QtWidgets.QWidget(self)
        # Create figure
        figure = matplotlib.figure.Figure()
        figure.suptitle(self._filename)
        canvas = backend.FigureCanvasQTAgg(figure)
        canvas.setParent(widget)
        axes = figure.add_subplot(1, 1, 1)
        toolbar = backend.NavigationToolbar2QT(canvas, widget)
        # Draw line
        if time is None:
            axes.plot(data)
        else:
            axes.plot(time, data)
        # Create a layout
        vbox = QtWidgets.QVBoxLayout()
        vbox.addWidget(canvas)
        vbox.addWidget(toolbar)
        widget.setLayout(vbox)
        self._figures.append(figure)
        self._axes.append(axes)
        return widget


class PatchMasterTab(SweepSourceTab):
    """ A widget displaying a PatchMaster series. """
    pass


class TxtTab(GraphTabWidget):
    """ A widget displaying a .txt file (with lots of heuristics!). """

    def __init__(self, parent, data, filename):
        super().__init__(parent)

        self._filename = filename

        # Find usable data
        if np.prod(data.shape) == np.max(data.shape):
            # 1d data
            data = data.reshape((np.max(data.shape),))
            # Create tab
            tab = self.create_graph_tab(None, data)
            self.addTab(tab, 'series 1')
        elif len(data.shape) == 2:
            # 2d data, assume longest axis is time
            if data.shape[0] < data.shape[1]:
                data = data.T
            # Check if first or last entry could be time
            if not np.any(data[0, 1:] < data[0, :-1]):
                time = data[0]
                data = data[1:]
            elif not np.any(data[-1, 1:] < data[-1, :-1]):
                time = data[-1]
                data = data[:-1]
            else:
                time = None
            # Create tabs
            for k, column in enumerate(data):
                tab = self.create_graph_tab(time, data)
                self.addTab(tab, 'series ' + str(1 + k))
        else:
            # Nothing that can be used? Show error
            QtWidgets.QMessageBox.critical(
                self, TITLE,
                'Unable to load file: unable to parse file contents.')
            return

    def create_graph_tab(self, time, data):
        """
        Creates a widget displaying a time series.
        """
        widget = QtWidgets.QWidget(self)
        # Create figure
        figure = matplotlib.figure.Figure()
        figure.suptitle(self._filename)
        canvas = backend.FigureCanvasQTAgg(figure)
        canvas.setParent(widget)
        axes = figure.add_subplot(1, 1, 1)
        toolbar = backend.NavigationToolbar2QT(canvas, widget)
        # Draw line
        if time is None:
            axes.plot(data)
        else:
            axes.plot(time, data)
        # Create a layout
        vbox = QtWidgets.QVBoxLayout()
        vbox.addWidget(canvas)
        vbox.addWidget(toolbar)
        widget.setLayout(vbox)
        self._figures.append(figure)
        self._axes.append(axes)
        return widget


class WcpTab(SweepSourceTab):
    pass




#!/usr/bin/env python3
#
# Tests the myokit._aux module.
#
# This file is part of Myokit.
# See http://myokit.org for copyright, sharing, and licensing details.
#
import io
import os
import re
import unittest

import myokit

from myokit.tests import DIR_DATA, WarningCollector


class AuxTest(unittest.TestCase):
    """
    Test various methods from myokit._aux.
    """

    def test_benchmarker(self):
        # Deprecated alias of myokit.tools.Benchmarker
        with WarningCollector() as c:
            b = myokit.Benchmarker()
        self.assertIn('`myokit.Benchmarker` is deprecated', c.text())
        self.assertIsInstance(b, myokit.tools.Benchmarker)

    def test_date(self):
        # Test date formatting method.
        import time
        for i in range(3):
            a = time.strftime(myokit.DATE_FORMAT)
            b = myokit.date()
            if a == b:
                break
        self.assertEqual(a, b)

    def test_default_protocol(self):
        # Test default_protocol()

        # Test default version
        protocol = myokit.default_protocol()
        self.assertTrue(isinstance(protocol, myokit.Protocol))
        self.assertEqual(protocol.head().period(), 1000)

        # Test adapting the time unit
        model = myokit.Model()
        t = model.add_component('c').add_variable('t')
        t.set_rhs(0)
        protocol = myokit.default_protocol(model)
        self.assertTrue(isinstance(protocol, myokit.Protocol))
        self.assertEqual(protocol.head().period(), 1000)

        t.set_binding('time')
        protocol = myokit.default_protocol(model)
        self.assertTrue(isinstance(protocol, myokit.Protocol))
        self.assertEqual(protocol.head().period(), 1000)

        t.set_unit('s')
        protocol = myokit.default_protocol(model)
        self.assertTrue(isinstance(protocol, myokit.Protocol))
        self.assertEqual(protocol.head().period(), 1)

        t.set_unit('ms')
        protocol = myokit.default_protocol(model)
        self.assertTrue(isinstance(protocol, myokit.Protocol))
        self.assertEqual(protocol.head().period(), 1000)

    def test_default_script(self):
        # Test default script

        # Test without a model
        script = myokit.default_script()
        self.assertTrue(isinstance(script, str))
        self.assertIn('run(1000)', script)

        # Test adapting the time unit
        model = myokit.Model()
        t = model.add_component('c').add_variable('t')
        t.set_rhs(0)
        script = myokit.default_script(model)
        self.assertTrue(isinstance(script, str))
        self.assertIn('run(1000)', script)

        t.set_binding('time')
        script = myokit.default_script(model)
        self.assertTrue(isinstance(script, str))
        self.assertIn('run(1000)', script)

        t.set_unit('s')
        script = myokit.default_script(model)
        self.assertTrue(isinstance(script, str))
        self.assertIn('run(1.0)', script)

        t.set_unit('ms')
        script = myokit.default_script(model)
        self.assertTrue(isinstance(script, str))
        self.assertIn('run(1000)', script)

        # Test plotting membrane potential
        v = model.get('c').add_variable('v')
        v.set_label('membrane_potential')
        script = myokit.default_script(model)
        self.assertTrue(isinstance(script, str))
        self.assertIn("var = 'c.v'", script)

        # TODO: Run with tiny model?

    def test_format_float_dict(self):
        # Test myokit.format_float_dict, which is deprecated
        d = {'one': 1, 'Definitely two': 2, 'Three-ish': 3.1234567}
        with WarningCollector() as c:
            x = myokit.format_float_dict(d).splitlines()
        self.assertIn('`myokit.format_float_dict` is deprecated', c.text())
        self.assertEqual(len(x), 3)
        self.assertEqual(x[0], 'Definitely two = 2')
        self.assertEqual(x[1], 'Three-ish      = 3.1234567')
        self.assertEqual(x[2], 'one            = 1')

    def test_format_path(self):
        # Deprecated alias of myokit.tools.format_path
        root = os.path.join(os.path.abspath('.'), 'a')
        a = os.path.join(root, 'b', 'c')
        b = os.path.join(os.path.abspath('.'), 'a')
        with WarningCollector() as c:
            x = myokit.format_path(a, b)
        self.assertIn('`myokit.format_path` is deprecated', c.text())
        self.assertEqual(x, myokit.tools.format_path(a, b))

    def test_model_comparison(self):
        # Test the model comparison class.

        m1 = os.path.join(DIR_DATA, 'beeler-1977-model-compare-a.mmt')
        m2 = os.path.join(DIR_DATA, 'beeler-1977-model-compare-b.mmt')
        m1 = myokit.load_model(m1)
        m2 = myokit.load_model(m2)

        with myokit.tools.capture() as capture:
            c = myokit.ModelComparison(m1, m2, live=True)

        differences = [
            '[x] Mismatched Meta property in model: "desc"',
            '[x] Mismatched Meta property in model: "name"',
            '[2] Missing Meta property in model: "author"',
            '[1] Missing Meta property in model: "extra"',
            '[x] Mismatched User function <f(1)>',
            '[1] Missing User function <g(1)>.',
            '[x] Mismatched Time variable: [1]<engine.time> [2]<engine.toim>',
            '[x] Mismatched Initial value for <ina.h>',
            '[x] Mismatched Initial value for <ina.j>',
            '[x] Mismatched State at position 5: [1]<isi.d> [2]<isiz.d>',
            '[x] Mismatched State at position 6: [1]<isi.f> [2]<isiz.f>',
            '[2] Missing state at position 7',
            '[x] Mismatched RHS <calcium.Cai>',
            '[2] Missing Variable <engine.time>',
            '[1] Missing Variable <engine.toim>',
            '[x] Mismatched RHS <ina.h.alpha>',
            '[x] Mismatched RHS <ina.j>',
            '[2] Missing Variable <ina.j.beta>',
            '[1] Missing Variable <ina.j.jeta>',
            '[2] Missing Component <isi>',
            '[x] Mismatched LHS <ix1.x1>',
            '[x] Mismatched RHS <ix1.x1>',
            '[2] Missing Variable <ix1.x1.alpha>',
            '[2] Missing Variable <ix1.x1.beta>',
            '[x] Mismatched RHS <membrane.C>',
            '[x] Mismatched RHS <membrane.i_ion>',
            '[x] Mismatched unit <membrane.V>',
            '[1] Missing Component <isiz>',
        ]

        live = [
            'Comparing:',
            '  [1] beeler-1977',
            '  [2] beeler-1977-with-differences',
        ] + differences + [
            'Done',
            '  ' + str(len(differences)) + ' differences found',
        ]

        caught_differences = set(capture.text().splitlines())
        live = set(live)
        self.assertEqual(live, caught_differences)
        differences = set(differences)
        caught_differences = set(c.text().splitlines())
        self.assertEqual(differences, caught_differences)

        # Test equality method
        self.assertFalse(c.equal())
        self.assertTrue(myokit.ModelComparison(m1, m1).equal())
        self.assertTrue(myokit.ModelComparison(m2, m2).equal())

        # Test len and iterator interface
        self.assertEqual(len(c), len(differences))
        self.assertEqual(len([x for x in c]), len(differences))

        # Test reverse is similar
        d = myokit.ModelComparison(m2, m1)
        self.assertEqual(len(c), len(d))

        # Test detection of missing time variable
        # Note: user function disappears when cloning
        m3 = m1.clone()
        m3.binding('time').set_binding(None)
        c = myokit.ModelComparison(m1, m3)
        self.assertEqual(
            c.text(),
            '[2] Missing User function <f(1)>\n'
            '[2] Missing Time variable <engine.time>')
        self.assertEqual(len(c), 2)
        self.assertEqual(len(c), len(myokit.ModelComparison(m3, m1)))

    def test_numpy_writer(self):
        # Test NumPy expression writer obtaining method.

        import myokit
        w = myokit.numpy_writer()
        import myokit.formats
        import myokit.formats.python
        self.assertIsInstance(w, myokit.formats.python.NumPyExpressionWriter)

        # Test custom name method for this writer
        e = myokit.parse_expression('5 + 3 * x')
        self.assertEqual(w.ex(e), '5.0 + 3.0 * x')

        # Test with unvalidated model (no unames set)
        m = myokit.Model()
        c = m.add_component('c')
        x = c.add_variable('x')
        x.set_rhs('5 + x')
        self.assertEqual(w.ex(x.rhs()), '5.0 + c_x')

    def test_python_writer(self):
        # Test Python expression writer obtaining method.

        import myokit
        w = myokit.python_writer()
        import myokit.formats
        import myokit.formats.python
        self.assertIsInstance(w, myokit.formats.python.PythonExpressionWriter)

        # Test custom name method for this writer
        e = myokit.parse_expression('5 + 3 * x')
        self.assertEqual(w.ex(e), '5.0 + 3.0 * x')

        # Test with unvalidated model (no unames set)
        m = myokit.Model()
        c = m.add_component('c')
        x = c.add_variable('x')
        x.set_rhs('5 + x')
        self.assertEqual(w.ex(x.rhs()), '5.0 + c_x')

    def test_run(self):
        # Test run() method.

        m, p, _ = myokit.load('example')
        x = '\n'.join([
            'import myokit',
            'm = get_model()',  # Test magic methods
            'p = get_protocol()',
        ])
        with myokit.tools.capture():
            myokit.run(m, p, x)
        with myokit.tools.capture():
            myokit.run(m, p, '[[script]]\n' + x)
        self.assertRaises(ZeroDivisionError, myokit.run, m, p, 'print(1 / 0)')

        # Test with stringio
        x = "print('Hi there')"
        s = io.StringIO()
        with myokit.tools.capture() as c:
            myokit.run(m, p, x, stderr=s, stdout=s)
        self.assertEqual(c.text(), '')
        self.assertEqual(s.getvalue(), 'Hi there\n')

    def test_step(self):
        # Test the step() method.
        m1 = myokit.load_model(os.path.join(DIR_DATA, 'beeler-1977-model.mmt'))

        # For whatever reason, CI gives slightly different final digits some
        # times.
        r1 = re.compile(r'[a-zA-Z]\w*\.[a-zA-Z]\w*\s+([^\s]+)\s+([^\s]+)')
        r2 = re.compile(r'\s+([0-9e.+-]+)')
        r3 = re.compile(r'^[ ^]*$')
        r4 = re.compile(r'^Found \([0-9]+\) small mismatches.$')

        def almost_equal(x, y):
            # Easy checks
            if x == y:
                return True

            # Get numbers with regex and compare
            # First case: assume two numbers on each line. 2nd is tested
            g1, g2 = r1.match(x), r1.match(y)
            if g1 is not None and g2 is not None:
                x, y = float(g1.group(2)), float(g2.group(2))
                return myokit.float.close(x, y)

            # Second case: one number per line.
            g1, g2 = r2.match(x), r2.match(y)
            if g1 is not None and g2 is not None:
                x, y = float(g1.group(1)), float(g2.group(1))
                return myokit.float.close(x, y)

            # Third case: "   ^^^ "
            # These don't matter: we've already seen that the floats are close
            g1, g2 = r3.match(x), r3.match(y)
            if g1 is not None and g2 is not None:
                return True

            # Fourth case: Found (x) small differences
            # These don't matter: we've already seen that the floats are close
            g1, g2 = r4.match(x), r4.match(y)
            if g1 is not None and g2 is not None:
                return True

            return x == y

        # Test simple output
        x = myokit.step(m1).splitlines()
        y = [
            'Evaluating state vector derivatives...',
            '-' * 79,
            'Name         Initial value             Derivative at t=0       ',
            '-' * 79,
            'membrane.V   -8.46219999999999999e+01  -3.97224086575331814e-04',
            'calcium.Cai   1.99999999999999991e-07  -1.56608433137725457e-09',
            'ina.m         1.00000000000000002e-02   7.48738392280519083e-02',
            'ina.h         9.89999999999999991e-01  -1.78891889478854579e-03',
            'ina.j         9.79999999999999982e-01  -3.06255006833574140e-04',
            'isi.d         3.00000000000000006e-03  -5.11993904291850035e-06',
            'isi.f         9.89999999999999991e-01   1.88374114688215870e-04',
            'ix1.x1        4.00000000000000019e-04  -3.21682814207918156e-07',
            '-' * 79,
        ]
        for a, b in zip(x, y):
            self.assertTrue(almost_equal(a, b))
        self.assertEqual(len(x), len(y))

        # Test with an initial state
        # For whatever reason, CI gives slightly different final 3 digits some
        # times.
        state = [-80, 1e-7, 0.1, 0.9, 0.9, 0.1, 0.9, 0.1]
        x = myokit.step(m1, initial=state).splitlines()
        y = [
            'Evaluating state vector derivatives...',
            '-' * 79,
            'Name         Initial value             Derivative at t=0       ',
            '-' * 79,
            'membrane.V   -8.00000000000000000e+01   1.41230788219242243e+00',
            'calcium.Cai   9.99999999999999955e-08   1.68235244574188927e-07',
            'ina.m         1.00000000000000006e-01  -5.12333452433218284e+00',
            'ina.h         9.00000000000000022e-01   1.30873415607557463e-02',
            'ina.j         9.00000000000000022e-01   1.43519283896554857e-03',
            'isi.d         1.00000000000000006e-01  -1.06388689494351027e-02',
            'isi.f         9.00000000000000022e-01   1.81759609957233962e-03',
            'ix1.x1        1.00000000000000006e-01  -4.72598388279933061e-03',
            '-' * 79,
        ]
        for a, b in zip(x, y):
            self.assertTrue(almost_equal(a, b))
        self.assertEqual(len(x), len(y))

        # Test comparison against another model (with both models the same)
        m2 = m1.clone()
        x = myokit.step(m1, reference=m2).splitlines()
        y = [
            'Evaluating state vector derivatives...',
            '-' * 79,
            'Name         Initial value             Derivative at t=0       ',
            '-' * 79,
            'membrane.V   -8.46219999999999999e+01  -3.97224086575331814e-04',
            '                                       -3.97224086575331814e-04',
            '',
            'calcium.Cai   1.99999999999999991e-07  -1.56608433137725457e-09',
            '                                       -1.56608433137725457e-09',
            '',
            'ina.m         1.00000000000000002e-02   7.48738392280519083e-02',
            '                                        7.48738392280519083e-02',
            '',
            'ina.h         9.89999999999999991e-01  -1.78891889478854579e-03',
            '                                       -1.78891889478854579e-03',
            '',
            'ina.j         9.79999999999999982e-01  -3.06255006833574140e-04',
            '                                       -3.06255006833574140e-04',
            '',
            'isi.d         3.00000000000000006e-03  -5.11993904291850035e-06',
            '                                       -5.11993904291850035e-06',
            '',
            'isi.f         9.89999999999999991e-01   1.88374114688215870e-04',
            '                                        1.88374114688215870e-04',
            '',
            'ix1.x1        4.00000000000000019e-04  -3.21682814207918156e-07',
            '                                       -3.21682814207918156e-07',
            '',
            'Model check completed without errors.',
            '-' * 79,
        ]
        for a, b in zip(x, y):
            self.assertTrue(almost_equal(a, b))
        self.assertEqual(len(x), len(y))

        # Test comparison against another model, with an initial state
        x = myokit.step(m1, reference=m2, initial=state).splitlines()
        y = [
            'Evaluating state vector derivatives...',
            '-' * 79,
            'Name         Initial value             Derivative at t=0       ',
            '-' * 79,
            'membrane.V   -8.00000000000000000e+01   1.41230788219242243e+00',
            '                                        1.41230788219242243e+00',
            '',
            'calcium.Cai   9.99999999999999955e-08   1.68235244574188927e-07',
            '                                        1.68235244574188927e-07',
            '',
            'ina.m         1.00000000000000006e-01  -5.12333452433218284e+00',
            '                                       -5.12333452433218284e+00',
            '',
            'ina.h         9.00000000000000022e-01   1.30873415607557463e-02',
            '                                        1.30873415607557463e-02',
            '',
            'ina.j         9.00000000000000022e-01   1.43519283896554857e-03',
            '                                        1.43519283896554857e-03',
            '',
            'isi.d         1.00000000000000006e-01  -1.06388689494351027e-02',
            '                                       -1.06388689494351027e-02',
            '',
            'isi.f         9.00000000000000022e-01   1.81759609957233962e-03',
            '                                        1.81759609957233962e-03',
            '',
            'ix1.x1        1.00000000000000006e-01  -4.72598388279933061e-03',
            '                                       -4.72598388279933061e-03',
            '',
            'Model check completed without errors.',
            '-' * 79,
        ]
        for a, b in zip(x, y):
            self.assertTrue(almost_equal(a, b))
        self.assertEqual(len(x), len(y))

        # Test comparison against stored data
        # And get all the different error outputs
        ref = m1.evaluate_derivatives()
        ref[0] += 5e-20  # Numerically indistinguishable
        ref[2] += 1e-16  # Tiny error
        ref[4] *= -1     # Sign error
        ref[6] += 2e-4   # Big error
        ref[7] *= 100    # Exponent

        x = myokit.step(m1, reference=ref).splitlines()
        y = [
            'Evaluating state vector derivatives...',
            '-' * 79,
            'Name         Initial value             Derivative at t=0       ',
            '-' * 79,
            'membrane.V   -8.46219999999999999e+01  -3.97224086575331814e-04',
            '                                       -3.97224086575331760e-04'
            ' <= 1 eps',
            '',
            'calcium.Cai   1.99999999999999991e-07  -1.56608433137725457e-09',
            '                                       -1.56608433137725457e-09',
            '',
            'ina.m         1.00000000000000002e-02   7.48738392280519083e-02',
            '                                        7.48738392280520054e-02'
            ' ~ 5.8 eps',
            '                                                      ^^^^^^^^^',
            'ina.h         9.89999999999999991e-01  -1.78891889478854579e-03',
            '                                       -1.78891889478854579e-03',
            '',
            'ina.j         9.79999999999999982e-01  -3.06255006833574140e-04',
            '                                        3.06255006833574140e-04'
            ' sign',
            '                                       ^^^^^^^^^^^^^^^^^^^^^^^^',
            'isi.d         3.00000000000000006e-03  -5.11993904291850035e-06',
            '                                       -5.11993904291850035e-06',
            '',
            'isi.f         9.89999999999999991e-01   1.88374114688215870e-04',
            '                                        3.88374114688215880e-04'
            ' X',
            '                                        ^^^^^^^^^^^^^^^^^^^^^^^',
            'ix1.x1        4.00000000000000019e-04  -3.21682814207918156e-07',
            '                                       -3.21682814207918156e-05'
            ' exponent',
            '                                                           ^^^^',
            'Found (3) large mismatches between output and reference values.',
            'Found (1) small mismatches.',
            '-' * 79,
        ]
        for a, b in zip(x, y):
            self.assertTrue(almost_equal(a, b))
        self.assertEqual(len(x), len(y))

        # Test positive/negative zero comparison
        m1 = myokit.Model()
        c = m1.add_component('c')
        t = c.add_variable('t')
        t.set_binding('time')
        t.set_rhs(0)
        x = c.add_variable('x')
        x.promote(1)
        x.set_rhs('-0.0')
        y = c.add_variable('y')
        y.promote(1)
        y.set_rhs('0.0')
        m2 = m1.clone()
        m2.get('c.x').set_rhs(0.0)

        x = myokit.step(m1, reference=m2).splitlines()
        y = [
            'Evaluating state vector derivatives...',
            '-' * 79,
            'Name  Initial value             Derivative at t=0       ',
            '-' * 79,
            'c.x    1.00000000000000000e+00  -0.00000000000000000e+00',
            '                                 0.00000000000000000e+00',
            '',
            'c.y    1.00000000000000000e+00   0.00000000000000000e+00',
            '                                 0.00000000000000000e+00',
            '',
            'Model check completed without errors.',
            '-' * 79,
        ]
        for a, b in zip(x, y):
            self.assertEqual(a, b)
        self.assertEqual(len(x), len(y))

    def test_strfloat(self):
        # Deprecated alias of myokit.float.str
        args = ['-1.234', True, myokit.SINGLE_PRECISION]
        with WarningCollector() as c:
            x = myokit.strfloat(*args)
        self.assertIn('`myokit.strfloat` is deprecated', c.text())
        self.assertEqual(x, myokit.float.str(*args))

    def test_time(self):
        # Test time formatting method.
        import time
        for i in range(6):
            a = time.strftime(myokit.TIME_FORMAT)
            b = myokit.time()
            if a == b:
                break
        self.assertEqual(a, b)

    def test_version(self):
        # Test the version() method.

        # Raw version
        raw = myokit.version(raw=True)
        self.assertIsInstance(raw, str)
        parts = raw.split('.')
        self.assertTrue(len(parts) in [3, 4])

        # Formatted version
        v = myokit.version()
        v = v.splitlines()
        self.assertEqual(len(v), 3)


if __name__ == '__main__':
    unittest.main()



#!/usr/bin/env python3
#
# Tests the Unit class
#
# This file is part of Myokit.
# See http://myokit.org for copyright, sharing, and licensing details.
#
import unittest

import myokit


class MyokitUnitTest(unittest.TestCase):

    def test_create(self):
        # Test basic unit creation.

        myokit.Unit.parse_simple('mV')
        myokit.Unit.parse_simple('g')
        myokit.Unit.parse_simple('kg')

        myokit.Unit([0, 0, 0, 0, 0, 0, 0])
        self.assertRaises(ValueError, myokit.Unit, [0, 0, 0, 0, 0, 0])
        self.assertRaises(ValueError, myokit.Unit, [0, 0, 0, 0, 0, 0, 0, 0])

    def test_can_convert(self):
        # Test :meth:`Unit.can_convert()`.

        self.assertTrue(myokit.Unit.can_convert(
            myokit.units.volt, myokit.units.mV))
        self.assertFalse(myokit.Unit.can_convert(
            myokit.units.volt, myokit.units.ampere))

    def test_close(self):
        # Test the "close" method, to see if units are highly similar

        # Test for base unit without a prefix
        m = myokit.units.m
        m_close = m * (1 - 1e-15)
        m_far = m * 2
        not_m = m_close / m

        self.assertTrue(myokit.Unit.close(m, m))
        self.assertTrue(myokit.Unit.close(m, m_close))
        self.assertTrue(myokit.Unit.close(m_close, m))
        self.assertTrue(myokit.Unit.close(m_close, m_close))
        self.assertFalse(myokit.Unit.close(m, m_far))
        self.assertFalse(myokit.Unit.close(m, not_m))

        # Test for base unit with a prefix.
        # (Note that e.g. mV actually doesn't have an "m" when represented in
        # the base units Myokit uses!)
        mm = myokit.Unit.parse_simple('mm')
        mm_close = mm * (1 - 1e-15)
        mm_far = mm * 2
        not_mm = mm_close / mm

        self.assertTrue(myokit.Unit.close(m, m))
        self.assertTrue(myokit.Unit.close(m, m_close))
        self.assertTrue(myokit.Unit.close(m_close, m))
        self.assertTrue(myokit.Unit.close(m_close, m_close))
        self.assertFalse(myokit.Unit.close(m, m_far))
        self.assertFalse(myokit.Unit.close(m, not_m))

        # Test tiny, but feasible, units
        pF2 = myokit.units.pF**2
        not_pF2 = pF2 * 1.0001
        self.assertTrue(myokit.Unit.close(pF2, pF2))
        self.assertTrue(myokit.Unit.close(not_pF2, not_pF2))
        self.assertFalse(myokit.Unit.close(not_pF2, pF2))
        self.assertFalse(myokit.Unit.close(pF2, not_pF2))

        # The next test is not a requirement, but tests the current behavior:
        u = myokit.units.m * 1.000000001
        self.assertTrue(myokit.Unit.close(u, myokit.units.m))

    def test_conversion_factor(self):
        # Test :meth:`Unit.conversion_factor()`.

        cf = myokit.Unit.conversion_factor
        q = myokit.Quantity
        u = myokit.parse_unit

        # Same units
        self.assertEqual(cf('V', 'V'), q(1))

        # Prefixed units
        self.assertEqual(cf('m', 'km'), q(0.001, '1 (1000)'))
        self.assertEqual(cf('km', 'm'), q(1000, '1 (0.001)'))
        self.assertEqual(cf('hm', 'm'), q(100, '1 (0.01)'))

        # Units with a multiplier
        cm_per_inch = cf('inches', 'cm')
        self.assertEqual(type(cm_per_inch), myokit.Quantity)
        self.assertEqual(cm_per_inch.value(), 2.54)
        self.assertEqual(cm_per_inch.unit(), u('cm/inches'))

        # Old unit: uA/cm^2
        # New unit: A/cm^2
        # To get from old to new, multiply by 1e-6 [1 (1e6)]
        # Because uA*(A/uA) = A,
        #   the unit for that factor is A/uA = 1/1e-6 = 1e6
        self.assertEqual(cf('uA/cm^2', 'A/cm^2'), q(1e-6, '1 (1e6)'))
        self.assertEqual(cf('uA/cm^2', 'A/cm^2'), q(1e-6, 'A/uA'))
        self.assertEqual(cf('uA/cm^2', 'A/cm^2'), q('1e-6 [A/uA]'))

        # Incompatible units
        self.assertRaises(
            myokit.IncompatibleUnitError, cf, u('g'), u('s'))
        self.assertRaises(
            myokit.IncompatibleUnitError, cf, u('N'), u('C'))

        # Convertible via helpers

        # Old unit: uA/cm^2
        # New unit: uA/uF
        # Helper: 2 [uF/cm^2]
        # To get from old to new, multiply by 0.5 [cm^2/uF]
        self.assertEqual(
            cf('uA/cm^2', 'uA/uF', ['0.5 [cm^2/uF]']), q(0.5, u('cm^2/uF')))
        self.assertEqual(
            cf('uA/cm^2', 'uA/uF', ['2 [uF/cm^2]']), q(0.5, u('cm^2/uF')))

        # Old unit: uA/uF = A/F = pA/pF
        # New unit: pA
        # Helper: 123 pF
        # To get from old to new, multiply by 123 [pF]
        self.assertEqual(
            cf('uA/uF', 'pA', ['123 [pF]']), q('123 [pF]'))
        self.assertEqual(
            cf('uA/uF', 'nA', ['123 [pF]']), q('0.123 [nF]'))
        self.assertEqual(
            cf('uA/cm^2', 'uA', ['0.123 [cm^2]']), q('0.123 [cm^2]'))

        # And:
        #   uA/cm^2 * cm^2 = uA
        #   uA/cm^2 * cm^2 * nA/uA = nA
        # So expecting 1 [cm^2] * 1000 [nA/uA] = 1000 [1 (1e-3)]
        # Factor in the 0.123 cm^2 to get 123 [cm^2 (1e-3)]
        self.assertEqual(
            cf('uA/cm^2', 'nA', ['0.123 [cm^2]']), q('123 [cm^2 (1e-3)]'))

        # Multiple factors to try
        h1 = ['123 [pF]', '0.123 [cm^2]']
        h2 = h1[::-1]
        self.assertEqual(cf('uA/uF', 'nA', h1), q('0.123 [nF]'))
        self.assertEqual(cf('uA/uF', 'nA', h2), q('0.123 [nF]'))
        self.assertEqual(cf('uA/cm^2', 'nA', h1), q('123 [cm^2 (1e-3)]'))
        self.assertEqual(cf('uA/cm^2', 'nA', h2), q('123 [cm^2 (1e-3)]'))

        # No good factors to try
        self.assertRaisesRegex(
            myokit.IncompatibleUnitError, 'even with', cf, 'mV', 'nA', h2)

    def test_convert(self):
        # Test :meth:`Unit.convert()`.

        # Test with float and int
        mV = myokit.units.mV
        V = myokit.units.V
        cv = myokit.Unit.convert
        self.assertEqual(cv(2, mV, V), 0.002)
        self.assertEqual(cv(2.0, V, mV), 2000)

        # Test with quantity
        q = myokit.Quantity
        self.assertEqual(cv(q(2, mV), mV, V), q(0.002, V))

        # None and dimensionless are ok
        d = myokit.units.dimensionless
        self.assertEqual(cv(1, None, None), 1)
        self.assertEqual(cv(1, d, None), 1)
        self.assertEqual(cv(1, None, d), 1)
        self.assertEqual(cv(1, d, d), 1)
        self.assertRaises(
            myokit.IncompatibleUnitError, myokit.Unit.convert, 1, d, V)
        self.assertRaises(
            myokit.IncompatibleUnitError, myokit.Unit.convert, 1, V, d)
        self.assertRaises(
            myokit.IncompatibleUnitError, myokit.Unit.convert, 1, None, V)
        self.assertRaises(
            myokit.IncompatibleUnitError, myokit.Unit.convert, 1, V, None)

        # Strings can be parsed
        self.assertEqual(cv(1, None, '1'), 1)
        self.assertEqual(cv(1, '1', None), 1)
        self.assertEqual(cv(1, 'V', V), 1)
        self.assertEqual(cv(1, V, 'V'), 1)
        self.assertRaisesRegex(
            myokit.IncompatibleUnitError, 'from',
            myokit.Unit.convert, 1, V, 'A')
        self.assertRaisesRegex(
            myokit.IncompatibleUnitError, 'from',
            myokit.Unit.convert, 1, 'A', V)

    def test_exponents(self):
        # Test Unit.exponents

        x = myokit.Unit()
        self.assertEqual(x.exponents(), [0] * 7)

        x = myokit.units.m ** -2
        self.assertEqual(x.exponents(), [0, -2, 0, 0, 0, 0, 0])

    def test_float(self):
        # Test :meth:`Unit.__float__()`.
        x = myokit.Unit()
        x *= 123
        self.assertAlmostEqual(float(x), 123)

        # Can't convert unless dimensionless (but with any multiplier)
        x *= myokit.units.V
        self.assertRaises(TypeError, float, x)

    def test_operators(self):
        # Test overloaded unit operators.

        # Test div
        d = myokit.Unit()
        self.assertEqual(d.exponents(), [0] * 7)
        d = d / myokit.units.m
        self.assertEqual(d.exponents(), [0, -1, 0, 0, 0, 0, 0])
        d = d / myokit.units.m
        self.assertEqual(d.exponents(), [0, -2, 0, 0, 0, 0, 0])
        d = d / d
        self.assertEqual(d.exponents(), [0, 0, 0, 0, 0, 0, 0])

        # Test mul
        d = myokit.Unit()
        self.assertEqual(d.exponents(), [0] * 7)
        d = d * myokit.units.s
        self.assertEqual(d.exponents(), [0, 0, 1, 0, 0, 0, 0])
        d = d * myokit.units.m
        self.assertEqual(d.exponents(), [0, 1, 1, 0, 0, 0, 0])
        d = d * d
        self.assertEqual(d.exponents(), [0, 2, 2, 0, 0, 0, 0])

        # Test pow
        d = myokit.Unit()
        self.assertEqual(d.exponents(), [0] * 7)
        d *= myokit.units.s
        d *= myokit.units.m
        self.assertEqual(d.exponents(), [0, 1, 1, 0, 0, 0, 0])
        d = d**3
        self.assertEqual(d.exponents(), [0, 3, 3, 0, 0, 0, 0])

        # Test rdiv and rmul (i.e. with non-units)
        d = myokit.Unit()
        d *= myokit.units.meter
        self.assertEqual(d._m, 0)
        self.assertEqual(d.exponents(), [0, 1, 0, 0, 0, 0, 0])
        d = 1000 * d
        self.assertEqual(d._m, 3)
        d = 1 / d
        self.assertEqual(d._m, -3)
        self.assertEqual(d.exponents(), [0, -1, 0, 0, 0, 0, 0])
        d = 100 * d
        self.assertEqual(d._m, -1)
        d = 10 * d
        self.assertEqual(d._m, 0)

    def test_multiplier(self):
        # Test :meth:`Unit.multiplier()` and :meth:`Unit.multiplier_log_10`.
        d = myokit.Unit()
        d *= myokit.units.meter
        d *= 1000
        self.assertEqual(d.multiplier(), 1000)
        self.assertEqual(d.multiplier_log_10(), 3)

    def test_parse_simple(self):
        # Test edge cases for :meth:`Unit.parse_simple()`.

        # Easy case
        self.assertEqual(myokit.Unit.parse_simple('mV'), myokit.units.mV)

        # Bad quantifier
        self.assertRaisesRegex(
            KeyError, 'Unknown quantifier', myokit.Unit.parse_simple, 'jV')

        # Not a quantifiable unit
        self.assertRaisesRegex(
            KeyError, 'cannot have quantifier', myokit.Unit.parse_simple,
            'mNewton')

        # Unknown unit
        self.assertRaisesRegex(
            KeyError, 'Unknown unit', myokit.Unit.parse_simple, 'Frog')

    def test_powers(self):
        # Fractional powers are allowed now

        m2 = myokit.units.m**2
        s15 = myokit.units.s**1.5
        mol123 = myokit.units.mol**1.23
        x = m2 * s15 / mol123
        self.assertEqual(x.exponents(), [0, 2, 1.5, 0, 0, 0, -1.23])

    def test_register_errors(self):
        # Test errors for Unit.register (rest is already used a lot).
        self.assertRaises(TypeError, myokit.Unit.register, 4, myokit.Unit())
        self.assertRaises(TypeError, myokit.Unit.register, 'hi', 4)

    def test_register_preferred_representation(self):
        # Test new representations can be registered

        u = myokit.units.m**8
        self.assertEqual(str(u), '[m^8]')
        try:
            myokit.Unit.register_preferred_representation(
                'abc', myokit.units.m**8)
            self.assertEqual(str(u), '[abc]')

            self.assertRaisesRegex(
                ValueError, 'must be a myokit.Unit',
                myokit.Unit.register_preferred_representation, 'x', 123)

        finally:
            # Bypassing the public API, this is bad test design!
            if u in myokit.Unit._preferred_representations:
                del myokit.Unit._preferred_representations[u]

    def test_str(self):
        # Test :meth:`Unit.str()`

        # Unit with representation in alternative base
        km_per_s = myokit.Unit([0, 1, -1, 0, 0, 0, 0], 3)

        # Myokit doesn't know km/s, it does know m/s so this should become:
        self.assertEqual(str(km_per_s), '[m/s (1000)]')

        # Myokit doesn't know MA/m^2
        mam2 = myokit.parse_unit('MA/m^2')
        self.assertEqual(str(mam2), '[A/m^2 (1e+06)]')

        # Simple units
        m = myokit.parse_unit('m')
        self.assertEqual(str(m), '[m]')
        im = 1 / m
        self.assertEqual(str(im), '[1/m]')

        # Predefined complex unit
        self.assertEqual(str(myokit.units.N), '[N]')

        # Low mulipliers
        um = m * 1e-6
        self.assertEqual(str(um), '[um]')
        um3 = myokit.parse_unit('um^3')
        self.assertEqual(str(um3), '[um^3]')
        attomol = myokit.parse_unit('amol')
        self.assertEqual(str(attomol), '[mol (1e-18)]')

        # High multipliers
        nn = myokit.units.N * 1e-2
        self.assertEqual(str(nn), '[N (0.01)]')
        nn = myokit.units.N * 1e2
        self.assertEqual(str(nn), '[N (100)]')
        nn = myokit.units.N * 1e7
        self.assertEqual(str(nn), '[N (1e+07)]')

        # Unit very similar to a known unit
        c = myokit.units.V
        d = c * 1.000000001
        self.assertFalse(c == d)
        self.assertTrue(myokit.Unit.close(c, d))
        self.assertEqual(str(c), '[V]')
        self.assertEqual(str(d), '[V]')

    def test_repr(self):
        # Test :meth:`Unit.repr()`.

        # Simple units
        m = myokit.parse_unit('m')
        self.assertEqual(repr(m), '[m]')
        im = 1 / m
        self.assertEqual(repr(im), '[1/m]')
        um = m * 1e-6

        # Predefined complex unit
        self.assertEqual(repr(myokit.units.N), '[g*m/s^2 (1000)]')

        # Low multipliers
        self.assertEqual(repr(um), '[m (1e-06)]')
        um3 = myokit.parse_unit('um^3')
        self.assertEqual(repr(um3), '[m^3 (1e-18)]')

        # High multipliers
        nn = myokit.units.N * 1e-2
        self.assertEqual(repr(nn), '[g*m/s^2 (10)]')
        nn = myokit.units.N * 1e7
        self.assertEqual(repr(nn), '[g*m/s^2 (1e+10)]')


if __name__ == '__main__':
    unittest.main()



#
# Exports model definitions to Latex files.
#
# This file is part of Myokit.
# See http://myokit.org for copyright, sharing, and licensing details.
#
import os

import myokit

from ._ewriter import LatexExpressionWriter


class PdfExporter(myokit.formats.Exporter):
    """
    This :class:`Exporter <myokit.formats.Exporter>` exports model equations to
    a simple latex document.
    """
    def __init__(self):
        super().__init__()

    def _clean(self, text):
        """
        Cleans some text for use in latex.
        """
        return text.replace('_', '\_')

    def post_export_info(self):
        return '\n'.join((
            'To create a pdf using pdflatex::',
            '',
            '  pdflatex filename.tex',
            '',
        ))

    def model(self, path, model, protocol=None):
        """
        Export to an xml document.
        """
        # Create an expression writer
        e = LatexExpressionWriter()

        # Write a simple file
        with open(path, 'w') as f:
            f.write('\\documentclass[fleqn]{article}\n')
            f.write('\\usepackage[a4paper,margin=0.7in]{geometry}\n')

            # Nice math stuff (\text{})
            f.write('\\usepackage{amsmath}\n')

            # Break lines automatically
            f.write('\\usepackage{breqn}\n')
            f.write('\\begin{document}\n')
            title = model.name() or 'Model'
            author = 'Dr. ' + self._clean(self.__class__.__name__)
            f.write('\\title{' + title + '}\n')
            f.write('\\author{' + author + '}\n')
            f.write('\\maketitle\n')

            # Introduction
            try:
                text = model.meta['desc']
                text = str(text)
                text.replace('\\', '\\\\')
                f.write('\\section{Introduction}\n')
                f.write(text)
            except KeyError:    # pragma: no cover
                # No need to test this
                pass

            # Initial conditions
            f.write('\\section{Initial conditions}\n')
            for v in model.states():
                f.write('\\begin{dmath}\n')
                f.write(e.ex(myokit.Name(v)))
                f.write(' = ')
                f.write(e.ex(v.initial_value()))
                f.write('\\end{dmath}\n')

            # Write each component
            for c in model.components():
                f.write('\\section{' + self._clean(c.name()) + '}\n')
                try:
                    text = c.meta['desc']
                    text = str(text)
                    text.replace('\\', '\\\\')
                    f.write(text)
                except KeyError:
                    pass
                for v in c.variables(deep=True):
                    f.write('\\begin{dmath}\n')
                    f.write(e.eq(v.eq()) + '\n')
                    f.write('\\end{dmath}\n')

            # End of document
            f.write('\\end{document}\n')

    def supports_model(self):
        """
        Returns ``True``
        """
        return True


class PosterExporter(myokit.formats.Exporter):
    """
    This :class:`Exporter <myokit.formats.Exporter>` exports model equations to
    a terrifying poster format, designed to strike fear into the heart of its
    beholders.
    """
    def __init__(self):
        super().__init__()

    def post_export_info(self):
        return '\n'.join((
            'To create a pdf using pdflatex::',
            '',
            '  pdflatex filename.tex',
            '',
            'This may require installing some latex packages.',
        ))

    def model(self, path, model, protocol=None):
        """
        Exports a model to an xml document.
        """
        path = os.path.abspath(os.path.expanduser(path))

        # Create an expression writer
        e = LatexExpressionWriter()

        # Write a simple file
        with open(path, 'w') as f:
            f.write('\\documentclass{article}\n')
            f.write('\\usepackage[a4paper,landscape,margin=0.7in]{geometry}\n')

            # Nice math stuff (\text{})
            f.write('\\usepackage{amsmath}\n')

            # Break lines automatically
            f.write('\\usepackage{breqn}\n')

            # Growing page
            f.write('\\usepackage[active,tightpage]{preview}\n')
            f.write('\\renewcommand{\\PreviewBorder}{1in}\n')
            f.write('\\begin{document}\n')
            f.write('\\begin{preview}\n')
            title = model.name() or 'Model'
            f.write('\\title{' + title + '}\n')
            f.write('\\author{}\n')
            f.write('\\date{}\n')
            f.write('\\maketitle\n')

            # Write each component
            for c in model.components():
                for v in c.variables(deep=True):
                    f.write('\\(\n')
                    f.write(e.eq(v.eq()) + '\n')
                    f.write('\\)\n')

            # End of document
            f.write('\\end{preview}\n')
            f.write('\\end{document}\n')

    def supports_model(self):
        """
        Returns ``True``.
        """
        return True



#!/usr/bin/env python3
#
# Tests the lib.markov module.
#
# This file is part of Myokit.
# See http://myokit.org for copyright, sharing, and licensing details.
#
import os
import unittest

import numpy as np

import myokit
import myokit.lib.markov as markov

from myokit.tests import DIR_DATA, WarningCollector


class LinearModelTest(unittest.TestCase):
    """
    Tests the linear model class.
    """

    def test_manual_creation(self):

        # Load model
        fname = os.path.join(DIR_DATA, 'clancy-1999-fitting.mmt')
        model = myokit.load_model(fname)

        # Select a number of states and parameters
        states = [
            'ina.C3',
            'ina.C2',
            'ina.C1',
            'ina.IF',
            'ina.IS',
            model.get('ina.O'),
        ]
        parameters = [
            'ina.p1',
            'ina.p2',
            'ina.p3',
            model.get('ina.p4'),
        ]
        current = 'ina.i'

        # Create a markov model
        m = markov.LinearModel(model, states, parameters, current)
        markov.LinearModel(model, states, parameters, model.get(current))

        # Check membrane potential
        self.assertEqual(m.membrane_potential(), 'membrane.V')

        # Check parameters
        parameters[3] = parameters[3].qname()
        for p in m.parameters():
            self.assertIn(p, parameters)

        # Test deprecated MarkovModel class
        with WarningCollector() as w:
            m2 = markov.MarkovModel(model, states, parameters, current)
        self.assertEqual(type(m2), markov.AnalyticalSimulation)
        self.assertIn('deprecated', w.text())

        # State doesn't exist
        self.assertRaisesRegex(
            markov.LinearModelError, 'Unknown state',
            markov.LinearModel, model, states + ['bert'], parameters, current)

        # State isn't a state
        self.assertRaisesRegex(
            markov.LinearModelError, 'not a state',
            markov.LinearModel, model, states + ['ina.i'], parameters, current)

        # State added twice
        self.assertRaisesRegex(
            markov.LinearModelError, 'twice',
            markov.LinearModel, model, states + ['ina.O'], parameters, current)

        # No parameters is allowed
        markov.LinearModel(model, states, None, current)

        # Non-literal parameter
        self.assertRaisesRegex(
            markov.LinearModelError, 'Unsuitable',
            markov.LinearModel, model, states, parameters + ['ina.i'], current)

        # Parameter added twice
        self.assertRaisesRegex(
            markov.LinearModelError, 'Parameter listed twice',
            markov.LinearModel,
            model, states, parameters + ['ina.p1'], current)

        # Unknown parameter
        self.assertRaisesRegex(
            markov.LinearModelError, 'Unknown parameter', markov.LinearModel,
            model, states, parameters + ['ina.p1000'], current)

        # Current is a state
        self.assertRaisesRegex(
            markov.LinearModelError, 'Current variable can not be a state',
            markov.LinearModel, model, states, parameters, 'ina.O')

        # Current is not a function of the states
        m2 = model.clone()
        markov.LinearModel(m2, states, parameters, current)
        m2.get(current).set_rhs(0)
        self.assertRaisesRegex(
            markov.LinearModelError, 'Current must be a function of',
            markov.LinearModel, m2, states, parameters, current)

        # Vm not given in model
        m2 = model.clone()
        markov.LinearModel(m2, states, parameters, current)
        m2.get('membrane.V').set_label(None)
        self.assertRaisesRegex(
            markov.LinearModelError, 'potential must be specified',
            markov.LinearModel, m2, states, parameters, current)
        markov.LinearModel(m2, states, parameters, current, vm='membrane.V')

        # Vm is a parameter
        self.assertRaisesRegex(
            markov.LinearModelError, 'list of parameters', markov.LinearModel,
            m2, states, parameters, current, vm=parameters[0])

        # Vm is a state
        self.assertRaisesRegex(
            markov.LinearModelError, 'list of states',
            markov.LinearModel, m2, states, parameters, current, vm=states[0])

        # Vm is the current
        self.assertRaisesRegex(
            markov.LinearModelError, 'be the current',
            markov.LinearModel, m2, states, parameters, current, vm=current)

        # States must have bidrectional dependencies
        m2 = model.clone()
        # Set bad config, but with columns still summing to zero
        m2.get('ina.C3').set_rhs('ina.C1')
        x = m2.get('ina.C2')
        x.set_rhs(str(x.rhs()) + '+ a12 * C2 + - b12 * C1')
        x = m2.get('ina.C1')
        x.set_rhs(str(x.rhs()) + ' - ina.C1')
        self.assertRaisesRegex(
            markov.LinearModelError, 'not vice versa',
            markov.LinearModel, m2, states, parameters, current)

        # States must sum to 1
        m2 = model.clone()
        m2.get(states[0]).set_initial_value(0.6)
        m2.get(states[1]).set_initial_value(0.6)
        self.assertRaisesRegex(
            markov.LinearModelError, 'sum of states',
            markov.LinearModel, m2, states, parameters, current)

        # Derivatives per column don't sum to zero
        m2 = model.clone()
        x = m2.get(states[0])
        x.set_rhs('2 * ' + str(x.rhs()))
        self.assertRaisesRegex(
            markov.LinearModelError, 'sum to non-zero',
            markov.LinearModel, m2, states, parameters, current)

        # Not a linear model
        m2 = model.clone()
        x = m2.get(states[0])
        y = m2.get(states[1])
        # Set rhs to something non-linear, make sure columns still sum to 0
        x.set_rhs(str(x.rhs()) + ' + C1^2')
        y.set_rhs(str(y.rhs()) + ' - C1^2')
        self.assertRaisesRegex(
            markov.LinearModelError, 'linear combination of states',
            markov.LinearModel, m2, states, parameters, current)

        # Current not a linear combination of states
        m2 = model.clone()
        m2.get(current).set_rhs('sqrt(ina.O)')
        self.assertRaisesRegex(
            markov.LinearModelError, 'linear combination of states',
            markov.LinearModel, m2, states, parameters, current)

    def test_linear_model_from_component(self):

        # Load model
        fname = os.path.join(DIR_DATA, 'clancy-1999-fitting.mmt')
        model = myokit.load_model(fname)

        # Create a markov model
        markov.LinearModel.from_component(model.get('ina'))

        # Test deprecated MarkovModel class
        with WarningCollector() as w:
            m = markov.MarkovModel.from_component(model.get('ina'))
        self.assertEqual(type(m), markov.AnalyticalSimulation)
        self.assertIn('deprecated', w.text())

        # Test partially automatic creation
        states = [
            'ina.C3',
            'ina.C2',
            'ina.C1',
            'ina.IF',
            'ina.IS',
            model.get('ina.O'),
        ]
        markov.LinearModel.from_component(model.get('ina'), states=states)

        parameters = [
            'ina.p1',
            'ina.p2',
            'ina.p3',
            model.get('ina.p4',)
        ]
        markov.LinearModel.from_component(
            model.get('ina'), parameters=parameters)

        current = 'ina.i'
        markov.LinearModel.from_component(model.get('ina'), current=current)

        markov.LinearModel.from_component(
            model.get('ina'), current=model.get(current))

        # No current --> This is allowed
        m2 = model.clone()
        m2.get('ina').remove_variable(m2.get('ina.i'))
        m = markov.LinearModel.from_component(m2.get('ina'))

        # Two currents
        m2 = model.clone()
        v = m2.get('ina').add_variable('i2')
        v.set_rhs(m2.get('ina.i').rhs().clone())
        self.assertRaisesRegex(
            markov.LinearModelError,
            'more than one variable that could be a current',
            markov.LinearModel.from_component, m2.get('ina'))

        # Explict vm
        m2 = model.clone()
        m2.get('membrane.V').set_label(None)
        markov.LinearModel.from_component(model.get('ina'), vm='membrane.V')
        self.assertRaisesRegex(
            markov.LinearModelError,
            'labeled as "membrane_potential"',
            markov.LinearModel.from_component, m2.get('ina'))

    def test_linear_model_matrices(self):

        # Create model
        fname = os.path.join(DIR_DATA, 'clancy-1999-fitting.mmt')
        model = myokit.load_model(fname)
        m = markov.LinearModel.from_component(model.get('ina'))

        # Test shape of output
        A, B = m.matrices(-20, range(21))
        self.assertEqual(A.shape, (6, 6))
        self.assertEqual(B.shape, (6, ))

        # Requires 21 parameters
        self.assertRaises(ValueError, m.matrices, -20, range(3))

    def test_linear_model_steady_state_1(self):
        # Test finding the steady-state of the Clancy model

        # Create model
        filename = os.path.join(DIR_DATA, 'clancy-1999-fitting.mmt')
        model = myokit.load_model(filename)
        m = markov.LinearModel.from_component(model.get('ina'))

        # Get steady state
        ss = np.array(m.steady_state())

        # Check that this is a valid steady state
        self.assertTrue(np.all(ss >= 0))
        self.assertTrue(np.all(ss <= 1))

        # Check that derivatives with ss are close to zero
        ss = list(ss)
        model.set_initial_values(ss + ss)    # Model has 2 ina's
        derivs = model.evaluate_derivatives()
        for i in range(len(ss)):
            self.assertAlmostEqual(0, derivs[i])

        # Try with awful parameters
        self.assertRaisesRegex(
            markov.LinearModelError, 'positive eigenvalues',
            m.steady_state, parameters=[-1] * 21)

    def test_linear_model_steady_state_2(self):
        # Test finding the steady-state of one of Dominic's models, which
        # exposed a bug in the steady state code

        # Create model
        filename = os.path.join(DIR_DATA, 'dom-markov.mmt')
        model = myokit.load_model(filename)
        m = markov.LinearModel.from_component(model.get('ikr'))

        # Get steady state
        ss = np.array(m.steady_state())

        # Check that this is a valid steady state
        self.assertTrue(np.all(ss >= 0))
        self.assertTrue(np.all(ss <= 1))

        # Check that derivatives with ss are close to zero
        model.set_initial_values(ss)
        derivs = model.evaluate_derivatives()
        for i in range(len(ss)):
            self.assertAlmostEqual(0, derivs[i])

    def test_rates(self):

        # Load model
        fname = os.path.join(DIR_DATA, 'clancy-1999-fitting.mmt')
        model = myokit.load_model(fname)

        # Create a markov model
        m = markov.LinearModel.from_component(model.get('ina'))

        # Test rates method runs
        self.assertEqual(len(m.rates()), 12)
        m.rates(parameters=[0.01] * 21)
        self.assertRaisesRegex(
            ValueError, 'Illegal parameter vector size',
            m.rates, parameters=[0.01] * 22)

    def test_initial_value_conversion(self):
        # Tests that initial value expressions are converted to floats

        fname = os.path.join(DIR_DATA, 'clancy-1999-fitting.mmt')
        model = myokit.load_model(fname)
        model.get('ina.C3').set_initial_value('1 / sqrt(7)')
        model.get('ina.C2').set_initial_value('-1 / log(ina_ref.p)')
        model.get('ina.C1').set_initial_value(0.5134619065149598)
        m = markov.LinearModel.from_component(model.get('ina'))
        x0 = m.default_state()
        self.assertEqual(len(x0), 6)
        self.assertIsInstance(x0[0], float)
        self.assertIsInstance(x0[1], float)
        self.assertAlmostEqual(x0[0], 0.3779644730092272)
        self.assertAlmostEqual(x0[1], 0.10857362047581297)


class AnalyticalSimulationTest(unittest.TestCase):
    """
    Tests :class:`myokit.lib.markov.AnalyticalSimulation`.
    """

    def test_create_and_run(self):
        # Test basics of analytical simulation

        # Create a simulation
        fname = os.path.join(DIR_DATA, 'clancy-1999-fitting.mmt')
        model = myokit.load_model(fname)
        m = markov.LinearModel.from_component(model.get('ina'))

        # Bad constructors
        self.assertRaisesRegex(
            ValueError, 'LinearModel', markov.AnalyticalSimulation, 1)
        self.assertRaisesRegex(
            ValueError, 'Protocol', markov.AnalyticalSimulation, m, 1)

        # Create properly
        s = markov.AnalyticalSimulation(m)

        # Times to evaluate at
        times = np.linspace(0, 100, 5)

        # Voltages to test at
        voltages = np.arange(-70, 0, 30)

        # Generate traces with "solve" method
        state = s.state()
        dstate = s.default_state()
        for v in voltages:
            s.set_membrane_potential(v)
            x, i = s.solve(times)

        # Solve shouldn't change the state
        self.assertEqual(state, s.state())
        self.assertEqual(dstate, s.default_state())

        # Run for a bit
        self.assertIsInstance(s.run(10), myokit.DataLog)

        # Calculate current for a particular state
        self.assertIsInstance(s.current(s.state()), float)

        # No current variable? Then current can't be calculated
        model2 = model.clone()
        model2.get('ina').remove_variable(model2.get('ina.i'))
        m2 = markov.LinearModel.from_component(model2.get('ina'))
        s2 = markov.AnalyticalSimulation(m2)
        self.assertRaisesRegex(
            Exception, 'did not specify a current', s2.current, s2.state())
        # But simulation still works
        self.assertIsInstance(s2.run(10), myokit.DataLog)
        del model2, m2, s2

        # Create protocol

        # Protocol times: prep, step, post, full
        tprep = 2800
        tstep = 15
        tpost = 0

        # Step voltages
        vhold = -80
        vlo = -140
        vhi = 100
        res = 50
        v = np.arange(vlo, vhi + res, res)
        p = myokit.pacing.steptrain(v, vhold, tprep, tstep, tpost)
        t = p.characteristic_time()

        # Create simulation with protocol (set_protocol is not supported)
        s = markov.AnalyticalSimulation(m, p)

        # Membrane potential and protocol can't be used simultaneously
        self.assertRaisesRegex(
            Exception, 'cannot be set if', s.set_membrane_potential, -80)

        # Pre should change the state and default state
        state = s.state()
        dstate = s.default_state()
        s.pre(tprep + tstep)
        self.assertNotEqual(state, s.state())
        self.assertNotEqual(dstate, s.default_state())
        self.assertRaises(ValueError, s.pre, -1)

        # Run should change the state, not the default state
        state = s.state()
        dstate = s.default_state()
        d = s.run(t)
        self.assertNotEqual(state, s.state())
        self.assertEqual(dstate, s.default_state())
        self.assertRaisesRegex(ValueError, 'Duration', s.run, -1)
        self.assertRaisesRegex(
            ValueError, 'Log interval', s.run, 1, log_interval=-1)
        d['hello'] = [1, 2, 3]
        self.assertRaisesRegex(ValueError, 'extra keys', s.run, 1, log=d)
        del d['hello']
        del d[next(iter(d.keys()))]
        self.assertRaisesRegex(ValueError, 'missing', s.run, 1, log=d)

        # Reset should reset the state
        s.reset()
        self.assertEqual(state, s.state())

        # Run can append to log
        d = s.run(10)
        n = len(d['engine.time'])
        e = s.run(1, log=d)
        self.assertIs(d, e)
        self.assertTrue(len(d['engine.time']) > n)

    def test_analytical_simulation_properties(self):
        # Test basic get/set methods of analytical simulation.

        # Create a simulation
        fname = os.path.join(DIR_DATA, 'clancy-1999-fitting.mmt')
        model = myokit.load_model(fname)
        m = markov.LinearModel.from_component(model.get('ina'))
        s = markov.AnalyticalSimulation(m)

        # membrane potential
        self.assertEqual(
            s.membrane_potential(), model.get('membrane.V').eval())
        s.set_membrane_potential(10)
        self.assertEqual(s.membrane_potential(), 10)

        # Parameter values
        p = list(range(len(s.parameters())))
        self.assertNotEqual(p, s.parameters())
        s.set_parameters(p)
        self.assertEqual(p, s.parameters())
        self.assertRaises(ValueError, s.set_parameters, p[:-1])

        # Change parameter with set_constant
        p[3] += 1
        self.assertNotEqual(p, s.parameters())
        s.set_constant(m.parameters()[3], p[3])
        self.assertEqual(p, s.parameters())

        # State
        state = np.zeros(len(s.state()))
        state[0] = 0.5
        state[1] = 0.5
        self.assertNotEqual(list(state), list(s.state()))
        s.set_state(state)
        self.assertEqual(list(state), list(s.state()))
        self.assertRaisesRegex(
            ValueError, 'Wrong size', s.set_state, state[:-1])
        state[0] += 0.1
        self.assertRaisesRegex(
            ValueError, 'sum to 1', s.set_state, state)
        state[0] = -.1
        state[1] = 1.1
        self.assertRaisesRegex(
            ValueError, 'negative', s.set_state, state)

        # Default state
        dstate = np.zeros(len(s.default_state()))
        dstate[0] = 0.5
        dstate[1] = 0.5
        self.assertNotEqual(list(dstate), list(s.default_state()))
        s.set_default_state(dstate)
        self.assertEqual(list(dstate), list(s.default_state()))
        self.assertRaisesRegex(
            ValueError, 'Wrong size', s.set_default_state, dstate[:-1])
        dstate[0] += 0.1
        self.assertRaisesRegex(
            ValueError, 'sum to 1', s.set_default_state, dstate)
        dstate[0] = -.1
        dstate[1] = 1.1
        self.assertRaisesRegex(
            ValueError, 'negative', s.set_default_state, dstate)

    def test_against_cvode(self):
        # Validate against a cvode sim.

        # Get a model
        fname = os.path.join(DIR_DATA, 'clancy-1999-fitting.mmt')
        model = myokit.load_model(fname)

        # Create a protocol
        vs = [-30, -20, -10]
        p = myokit.pacing.steptrain(
            vsteps=vs,
            vhold=-120,
            tpre=8,
            tstep=2,
            tpost=0)
        t = p.characteristic_time()

        # Run an analytical simulation
        dt = 0.01
        m = markov.LinearModel.from_component(model.get('ina'))
        s1 = markov.AnalyticalSimulation(m, p)
        d1 = s1.run(t, log_interval=dt).npview()

        s2 = myokit.Simulation(model, p)
        s2.set_tolerance(1e-8, 1e-8)
        d2 = s2.run(t, log_interval=dt).npview()

        # Test protocol output is the same
        e = np.abs(d1['membrane.V'] - d2['membrane.V'])
        if False:
            import matplotlib.pyplot as plt
            plt.figure()
            plt.plot(d1['membrane.V'])
            plt.plot(d2['membrane.V'])
            plt.show()
        self.assertEqual(np.max(e), 0)

        # Test current output is very similar
        e = np.abs(d1['ina.i'] - d2['ina.i'])
        if False:
            import matplotlib.pyplot as plt
            plt.figure()
            plt.plot(d1['ina.i'])
            plt.plot(d2['ina.i'])
            plt.figure()
            plt.plot(d1['ina.i'] - d2['ina.i'])
            plt.show()
        self.assertLess(np.max(e), 2e-4)


class DiscreteSimulationTest(unittest.TestCase):
    """
    Tests :class:`myokit.lib.markov.DiscreteSimulationTest`.
    """

    def test_basics(self):
        # Test the DiscreteSimulation class, running, resetting etc..

        # Create a simulation
        fname = os.path.join(DIR_DATA, 'clancy-1999-fitting.mmt')
        model = myokit.load_model(fname)
        m = markov.LinearModel.from_component(model.get('ina'))

        # Bad constructors
        self.assertRaisesRegex(
            ValueError, 'LinearModel', markov.DiscreteSimulation, 1)
        self.assertRaisesRegex(
            ValueError, 'Protocol', markov.DiscreteSimulation, m, 1)
        self.assertRaisesRegex(
            ValueError, 'at least 1',
            markov.DiscreteSimulation, m, nchannels=0)

        # Test running without a protocol
        s = markov.DiscreteSimulation(m)
        s.run(1)

        # Rest running for a very short time doesn't cause crash
        s.run(0)

        # Create protocol

        # Protocol times: prep, step, post, full
        tprep = 10
        tstep = 150
        tpost = 0

        # Step voltages
        vhold = -80
        vlo = -140
        vhi = 100
        res = 50
        v = np.arange(vlo, vhi + res, res)
        p = myokit.pacing.steptrain(v, vhold, tprep, tstep, tpost)

        # Create simulation with protocol (set_protocol is not supported)
        np.random.seed(1)
        s = markov.DiscreteSimulation(m, p)

        # Membrane potential and protocol can't be used simultaneously
        self.assertRaisesRegex(
            Exception, 'cannot be set if', s.set_membrane_potential, -80)

        # Pre should change the state and default state
        state = s.state()
        dstate = s.default_state()
        s.pre(tprep + tstep)
        self.assertNotEqual(state, s.state())
        self.assertNotEqual(dstate, s.default_state())
        self.assertRaisesRegex(ValueError, 'negative', s.pre, -1)

        # Run should change the state, not the default state
        state = s.state()
        dstate = s.default_state()
        d = s.run(15)
        self.assertNotEqual(state, s.state())
        self.assertEqual(dstate, s.default_state())
        self.assertRaisesRegex(ValueError, 'negative', s.run, -1)

        # Reset should reset the state
        s.reset()
        self.assertEqual(state, s.state())

        # Run can append to log
        n = len(d['engine.time'])
        e = s.run(1, log=d)
        self.assertIs(d, e)
        self.assertTrue(len(d['engine.time']) > n)
        self.assertEqual(len(d['engine.time']), len(d['membrane.V']))
        self.assertEqual(len(d['engine.time']), len(d['ina.i']))
        self.assertEqual(len(d['engine.time']), len(d['ina.O']))
        d2 = d.clone()
        del d2[next(iter(d2.keys()))]
        self.assertRaisesRegex(ValueError, 'missing', s.run, 1, log=d2)
        d2 = d.clone()
        d2['hello'] = [1, 2, 3]
        self.assertRaisesRegex(ValueError, 'extra', s.run, 1, log=d2)

        #
        # Test without current variable
        #
        model.get('ina').remove_variable(model.get('ina.i'))
        m = markov.LinearModel.from_component(model.get('ina'))

        # Create simulation with protocol (set_protocol is not supported)
        np.random.seed(1)
        s = markov.DiscreteSimulation(m, p)
        d = s.run(10)
        n = len(d['engine.time'])
        e = s.run(1, log=d)
        self.assertIs(d, e)
        self.assertTrue(len(d['engine.time']) > n)
        self.assertEqual(len(d['engine.time']), len(d['membrane.V']))
        self.assertEqual(len(d['engine.time']), len(d['ina.O']))
        self.assertNotIn('ina.i', d)
        d2 = d.clone()
        del d2[next(iter(d2.keys()))]
        self.assertRaisesRegex(ValueError, 'missing', s.run, 1, log=d2)
        d2 = d.clone()
        d2['hello'] = [1, 2, 3]
        self.assertRaisesRegex(ValueError, 'extra', s.run, 1, log=d2)

    def test_discrete_simulation_properties(self):
        # Test basic get/set methods of discrete simulation.

        # Create a simulation
        fname = os.path.join(DIR_DATA, 'clancy-1999-fitting.mmt')
        model = myokit.load_model(fname)
        m = markov.LinearModel.from_component(model.get('ina'))
        s = markov.DiscreteSimulation(m, nchannels=50)

        # membrane potential
        self.assertEqual(
            s.membrane_potential(), model.get('membrane.V').eval())
        s.set_membrane_potential(10)
        self.assertEqual(s.membrane_potential(), 10)

        # Number of channels
        self.assertEqual(s.number_of_channels(), 50)

        # Parameter values
        p = list(range(len(s.parameters())))
        self.assertNotEqual(p, s.parameters())
        s.set_parameters(p)
        self.assertEqual(p, s.parameters())
        self.assertRaises(ValueError, s.set_parameters, p[:-1])

        # Change parameter with set_constant
        p[3] += 1
        self.assertNotEqual(p, s.parameters())
        s.set_constant(m.parameters()[3], p[3])
        self.assertEqual(p, s.parameters())

        # State
        state = np.zeros(len(s.state()))
        state[0] = 25
        state[1] = 25
        self.assertNotEqual(list(state), list(s.state()))
        s.set_state(state)
        self.assertEqual(list(state), list(s.state()))
        self.assertRaisesRegex(
            ValueError, 'Wrong size', s.set_state, state[:-1])
        state[0] += 1
        self.assertRaisesRegex(
            ValueError, 'must equal', s.set_state, state)
        state[0] = -1
        state[1] = 51
        self.assertRaisesRegex(
            ValueError, 'negative', s.set_state, state)

        # Default state
        dstate = np.zeros(len(s.default_state()))
        dstate[0] = 25
        dstate[1] = 25
        self.assertNotEqual(list(dstate), list(s.default_state()))
        s.set_default_state(dstate)
        self.assertEqual(list(dstate), list(s.default_state()))
        self.assertRaisesRegex(
            ValueError, 'Wrong size', s.set_default_state, dstate[:-1])
        dstate[0] += 1
        self.assertRaisesRegex(
            ValueError, 'must equal', s.set_default_state, dstate)
        dstate[0] = -1
        dstate[1] = 51
        self.assertRaisesRegex(
            ValueError, 'negative', s.set_default_state, dstate)

        # Discretize state
        self.assertEqual(s.discretize_state([0.4, 0.6]), [20, 30])
        self.assertRaisesRegex(
            ValueError, 'must equal 1', s.discretize_state, [0.5, 0.6])


class MarkovFunctionsTest(unittest.TestCase):
    """
    Test cases for finding Markov models.
    """

    def test_convert_markov_models_to_compact_form(self):
        # Tests convert_markov_models_to_compact_form()

        # Load clancy model, has two versions of same markov model in it
        fname = os.path.join(DIR_DATA, 'clancy-1999-fitting.mmt')
        model1 = myokit.load_model(fname)

        models = markov.find_markov_models(model1)
        self.assertEqual(len(models), 2)
        m1, m2 = models

        # Check both models are full ODE form
        n1 = sum([1 for x in m1 if x.is_state()])
        self.assertEqual(n1, len(m1))
        n2 = sum([1 for x in m2 if x.is_state()])
        self.assertEqual(n2, len(m2))

        # Convert both compact form
        model2 = markov.convert_markov_models_to_compact_form(model1)
        models = markov.find_markov_models(model2)
        self.assertEqual(len(models), 2)
        m1, m2 = models
        n1 = sum([1 for x in m1 if x.is_state()])
        self.assertEqual(n1, len(m1) - 1)
        n2 = sum([1 for x in m2 if x.is_state()])
        self.assertEqual(n2, len(m2) - 1)

        # Check states evaluate to the same value
        self.assertEqual(
            model1.get('ina.C1').eval(), model2.get('ina.C1').eval())
        self.assertEqual(
            model1.get('ina.C2').eval(), model2.get('ina.C2').eval())
        self.assertEqual(
            model1.get('ina.C3').eval(), model2.get('ina.C3').eval())
        self.assertEqual(
            model1.get('ina.IF').eval(), model2.get('ina.IF').eval())
        self.assertEqual(
            model1.get('ina.IS').eval(), model2.get('ina.IS').eval())
        self.assertEqual(
            model1.get('ina.O').eval(), model2.get('ina.O').eval())

        # Doing it twice should have no effect
        model3 = markov.convert_markov_models_to_compact_form(model2)
        self.assertEqual(model2.code(), model3.code())

    def test_convert_markov_models_to_full_ode_form(self):
        # Tests convert_markov_models_to_compact_form()

        # Load clancy model, has two versions of same markov model in it
        fname = os.path.join(DIR_DATA, 'clancy-1999-fitting.mmt')
        model1 = myokit.load_model(fname)

        # Convert to compact form, and check that it worked
        model1 = markov.convert_markov_models_to_compact_form(model1)
        m1, m2 = markov.find_markov_models(model1)
        n1 = sum([1 for x in m1 if x.is_state()])
        self.assertEqual(n1, len(m1) - 1)
        n2 = sum([1 for x in m2 if x.is_state()])
        self.assertEqual(n2, len(m2) - 1)

        # Now convert to full form
        model2 = markov.convert_markov_models_to_full_ode_form(model1)
        m1, m2 = markov.find_markov_models(model2)
        n1 = sum([1 for x in m1 if x.is_state()])
        self.assertEqual(n1, len(m1))
        n2 = sum([1 for x in m2 if x.is_state()])
        self.assertEqual(n2, len(m2))

        # Check states evaluate to the same value
        self.assertEqual(
            model1.get('ina.C1').eval(), model2.get('ina.C1').eval())
        self.assertEqual(
            model1.get('ina.C2').eval(), model2.get('ina.C2').eval())
        self.assertEqual(
            model1.get('ina.C3').eval(), model2.get('ina.C3').eval())
        self.assertEqual(
            model1.get('ina.IF').eval(), model2.get('ina.IF').eval())
        self.assertEqual(
            model1.get('ina.IS').eval(), model2.get('ina.IS').eval())
        self.assertEqual(
            model1.get('ina.O').eval(), model2.get('ina.O').eval())

        # Doing it twice should have no effect
        model3 = markov.convert_markov_models_to_full_ode_form(model2)
        self.assertEqual(model2.code(), model3.code())

    def test_find_markov_models(self):
        # Tests find_markov_models()

        # Load clancy model, has two versions of same markov model in it
        fname = os.path.join(DIR_DATA, 'clancy-1999-fitting.mmt')
        model = myokit.load_model(fname)

        models = markov.find_markov_models(model)
        self.assertEqual(len(models), 2)

        # Check states and ordering
        m1, m2 = models
        self.assertEqual([v.qname() for v in m1], [
            'ina.C1', 'ina.C2', 'ina.C3', 'ina.IF', 'ina.IS', 'ina.O'])
        self.assertEqual([v.qname() for v in m2], [
            'ina_ref.C1', 'ina_ref.C2', 'ina_ref.C3', 'ina_ref.IF',
            'ina_ref.IS', 'ina_ref.O'])
        del models, m1, m2

        # Try with `1 - sum(xi)` state
        c = model.get('ina_ref')
        v = c.get('C3')
        v.demote()
        v.set_rhs('1 - C1 - C2 - IF - IS - O')
        models = markov.find_markov_models(model)
        self.assertEqual(len(models), 2)
        m1, m2 = models
        self.assertEqual([v.qname() for v in m2], [
            'ina_ref.C1', 'ina_ref.C2', 'ina_ref.C3', 'ina_ref.IF',
            'ina_ref.IS', 'ina_ref.O'])
        del models, m1, m2

        # Try with `1 - sum(xi)` state, with a funny RHS
        c = model.get('ina_ref')
        v = c.get('C3')
        v.set_rhs('-(+IF + C1 -(-IS - C2)) + 1 - O')
        models = markov.find_markov_models(model)
        self.assertEqual(len(models), 2)
        m1, m2 = models
        self.assertEqual([v.qname() for v in m2], [
            'ina_ref.C1', 'ina_ref.C2', 'ina_ref.C3', 'ina_ref.IF',
            'ina_ref.IS', 'ina_ref.O'])

    def test_find_markov_models_bad(self):
        # Tests find_markov_models() for non-markov models

        # Load clancy model, has two versions of same markov model in it
        fname = os.path.join(DIR_DATA, 'clancy-1999-fitting.mmt')
        moodel = myokit.load_model(fname)

        # Remove ina_ref component
        c = moodel.get('ina_ref')
        for v in c.variables(deep=True):
            v.set_rhs(0)
        for v in list(c.variables(deep=True)):
            c.remove_variable(v, recursive=True)

        # Only one markov model left at this point
        self.assertEqual(len(markov.find_markov_models(moodel)), 1)

        # Check searching states that no-one refers to
        # (And for cases where not each state is a linear combo)
        m = moodel.clone()
        for v in list(m.get('ina.C1').refs_by(True)):
            v.set_rhs(3)
        self.assertEqual(len(markov.find_markov_models(m)), 0)

        # Test with one 1-minus state
        m = moodel.clone()
        v = m.get('ina.C3')
        v.demote()
        v.set_rhs('1 - C1 - C2 - IF - IS - O')
        self.assertEqual(len(markov.find_markov_models(m)), 1)

        # Test without a 1
        v.set_rhs('2 - C1 - C2 - IF - IS - O')
        self.assertEqual(len(markov.find_markov_models(m)), 0)
        v.set_rhs('-C1 - C2 - IF - IS - O')
        self.assertEqual(len(markov.find_markov_models(m)), 0)

        # Test 1-... contains non-linear terms
        v.set_rhs('1 - C1 - C2 - IF - IS - O - O^2')
        self.assertEqual(len(markov.find_markov_models(m)), 0)

        # Test 1-... contains terms with a factor other than -1
        v.set_rhs('1 - C1 - C2 - IF - IS - O - O')
        self.assertEqual(len(markov.find_markov_models(m)), 0)

        # Test if there's multiple variables with a 1-... RHS
        m = moodel.clone()
        v = m.get('ina.C3')
        v.demote()
        v.set_rhs('1 - C1 - C2 - IF - IS - O')
        self.assertEqual(len(markov.find_markov_models(m)), 1)
        v = m.get('ina').add_variable('C4')
        v.set_rhs('1 - C1 - C2 - IF - IS - O')
        self.assertEqual(len(markov.find_markov_models(m)), 0)

        # But having an extra variable is fine, if the rest checks out!
        m = moodel.clone()
        v = m.get('ina').add_variable('C4')
        v.set_rhs('1 - C1 - C2 - C3 - IF - IS - O')
        self.assertEqual(len(markov.find_markov_models(m)), 1)

        # Must have at least two states
        m = myokit.Model()
        c = m.add_component('c')
        t = c.add_variable('time')
        t.set_binding('time')
        v = c.add_variable('v')
        v.promote(0.1)
        self.assertEqual(len(markov.find_markov_models(m)), 0)

    def test_linear_combination(self):
        # Tests _linear_combination()

        # Load model, to create interesting RHS
        fname = os.path.join(DIR_DATA, 'clancy-1999-fitting.mmt')
        model = myokit.load_model(fname)
        v1 = model.get('ina.C1')
        v2 = model.get('ina.C2')
        v3 = model.get('ina.C3')
        v4 = model.get('ina.IF')
        v5 = model.get('ina.IS')
        v6 = model.get('ina.O')

        # Test C1 rhs
        f = markov._linear_combination(v1.rhs(), [v1, v2, v3, v4, v5, v6])
        self.assertEqual(f[0].code(), '-(ina.a13 + ina.b12 + ina.b3)')
        self.assertEqual(f[1].code(), 'ina.a12')
        self.assertIsNone(f[2])
        self.assertEqual(f[3].code(), 'ina.a3')
        self.assertIsNone(f[4])
        self.assertEqual(f[5].code(), 'ina.b13')

        # Test double appearances
        v1.set_rhs('2 * C1 - 3 * C1 + C1 * sqrt(O) + C2 * IF')
        f = markov._linear_combination(v1.rhs(), [v1, v2, v3])
        self.assertEqual(f[0].code(), '2 + -3 + sqrt(ina.O)')
        self.assertEqual(f[1].code(), 'ina.IF')
        self.assertIsNone(f[2])

    def test_split_factor(self):
        # Tests _split_factor

        # Load model, to create interesting RHS
        fname = os.path.join(DIR_DATA, 'clancy-1999-fitting.mmt')
        model = myokit.load_model(fname)
        v1 = model.get('ina.C1')
        v2 = model.get('ina.C2')
        v3 = model.get('ina.C3')
        v4 = model.get('ina.IF')

        # Test simplest cases
        v4.set_rhs('C1')
        self.assertEqual(
            markov._split_factor(v4.rhs(), [v1]),
            (myokit.Name(v1), myokit.Number(1)))
        v4.set_rhs('+++C1')
        self.assertEqual(
            markov._split_factor(v4.rhs(), [v1]),
            (myokit.Name(v1), myokit.Number(1)))
        v4.set_rhs('--C1')
        self.assertEqual(
            markov._split_factor(v4.rhs(), [v1]),
            (myokit.Name(v1), myokit.Number(1)))
        v4.set_rhs('---C1')
        self.assertEqual(
            markov._split_factor(v4.rhs(), [v1]),
            (myokit.Name(v1), myokit.PrefixMinus(myokit.Number(1))))

        # Test multiplication
        v4.set_rhs('C1 * 3')
        self.assertEqual(
            markov._split_factor(v4.rhs(), [v1]),
            (myokit.Name(v1), myokit.Number(3)))
        v4.set_rhs('C1 * (sqrt(C2) + C3)')
        self.assertEqual(
            markov._split_factor(v4.rhs(), [v1]),
            (myokit.Name(v1),
             myokit.Plus(myokit.Sqrt(myokit.Name(v2)), myokit.Name(v3))))

        # Test division
        v4.set_rhs('C1 / (sqrt(C2) + C3)')
        self.assertEqual(
            markov._split_factor(v4.rhs(), [v1]),
            (myokit.Name(v1),
             myokit.Divide(
                myokit.Number(1),
                myokit.Plus(myokit.Sqrt(myokit.Name(v2)), myokit.Name(v3)))))

        # Test division that's not allowed
        v4.set_rhs('(sqrt(C2) + C3) / C1')
        self.assertRaisesRegex(
            ValueError, r'Non-linear function \(division\)',
            markov._split_factor, v4.rhs(), [v1])

        # Test with list of variables
        v4.set_rhs('C2 * 3')
        self.assertEqual(
            markov._split_factor(v4.rhs(), [v1, v2, v3]),
            (myokit.Name(v2), myokit.Number(3)))

        # Multiple variables is not allowed
        v4.set_rhs('C2 * C1')
        self.assertRaisesRegex(
            ValueError, 'must reference exactly one variable',
            markov._split_factor, v4.rhs(), [v1, v2, v3])

        # Zero variables is not allowed
        v4.set_rhs('C3')
        self.assertRaisesRegex(
            ValueError, 'must reference exactly one variable',
            markov._split_factor, v4.rhs(), [v1, v2])

        # Non-linear term is not allowed
        v4.set_rhs('sqrt(C1)')
        self.assertRaisesRegex(
            ValueError, 'Non-linear function',
            markov._split_factor, v4.rhs(), [v1, v2])

        # Multiple terms is not allowed
        v4.set_rhs('C2 - C2')
        self.assertRaisesRegex(
            ValueError, 'must be a single term',
            markov._split_factor, v4.rhs(), [v1, v2, v3])

    def test_split_terms(self):
        # Tests _split_terms

        # Load model, get rhs with lots of terms
        fname = os.path.join(DIR_DATA, 'clancy-1999-fitting.mmt')
        model = myokit.load_model(fname)

        # Simple case
        v1 = model.get('ina.C1')
        v2 = model.get('ina.C2')
        v3 = model.get('ina.C3')
        v4 = model.get('ina.IF')
        v5 = model.get('ina.IS')
        v6 = model.get('ina.O')
        v3.set_rhs('1 - C1 - C2 - IF - IS - O')
        terms = markov._split_terms(v3.rhs())
        self.assertEqual(terms[0], myokit.Number(1))
        self.assertEqual(terms[1], myokit.PrefixMinus(myokit.Name(v1)))
        self.assertEqual(terms[2], myokit.PrefixMinus(myokit.Name(v2)))
        self.assertEqual(terms[3], myokit.PrefixMinus(myokit.Name(v4)))
        self.assertEqual(terms[4], myokit.PrefixMinus(myokit.Name(v5)))
        self.assertEqual(terms[5], myokit.PrefixMinus(myokit.Name(v6)))
        del terms

        # Case with brackets
        v3.set_rhs('-(+(IF) + C1 -(-IS - C2)) + 1 - O')
        terms = markov._split_terms(v3.rhs())
        self.assertEqual(terms[0], myokit.PrefixMinus(myokit.Name(v4)))
        self.assertEqual(terms[1], myokit.PrefixMinus(myokit.Name(v1)))
        self.assertEqual(terms[2], myokit.PrefixMinus(myokit.Name(v5)))
        self.assertEqual(terms[3], myokit.PrefixMinus(myokit.Name(v2)))
        self.assertEqual(terms[4], myokit.Number(1))
        self.assertEqual(terms[5], myokit.PrefixMinus(myokit.Name(v6)))

        # Empty case
        v3.set_rhs('1 * C1 * C2')
        terms = markov._split_terms(v3.rhs())
        self.assertEqual(len(terms), 1)
        self.assertEqual(terms[0], v3.rhs())


if __name__ == '__main__':
    unittest.main()



#
# Imports step protocols from HEKA PatchMaster files.
#
# This file is part of Myokit.
# See http://myokit.org for copyright, sharing, and licensing details.
#
import myokit.formats


class PatchMasterImporter(myokit.formats.Importer):
    """
    This :class:`Importer <myokit.formats.Importer>` can import (step)
    protocols from "series" inside HEKA PatchMaster files.
    """

    def supports_protocol(self):
        return True

    def protocol(self, filename, group=None, series=None):
        """
        Attempts to load the protocol from the file at ``filename``.

        Because PatchMaster files can contain several experiments, a further
        selection can be made with the arguments:

        ``filename``
            The file to read
        ``group``
            The group to read from (as a string).
        ``series``
            The integer index of the desired series in the specified group.

        """
        from myokit.formats.heka import PatchMasterFile
        with PatchMasterFile(filename) as f:
            group = f.group(group)
            series = group[series]
            return series.stimulus().protocol()




#
# A tiny templating engine using a php style syntax.
#
# This file is part of Myokit.
# See http://myokit.org for copyright, sharing, and licensing details.
#
import ast
import io
import re
import sys
import traceback


class TemplateEngine:
    """
    A tiny templating engine using a php style syntax.

    Not intended for use in websites or with untrusted templates.

    Basic syntax::

        Hello <? print("world") ?>

    Fast syntax to write expressions::

        Hello <?= "world" ?>

    All processed template data is printed to the standard output stream or a
    stream specified by the user.
    """
    def __init__(self):
        super().__init__()
        self.stream = None
        self.error = None

    def error_details(self):
        """
        After a PypeError has been thrown, calling this will method will return
        a detailed (multi-line) error message.

        If no PypeError occurred the return value will be ``None``.
        """
        return self.error

    def process(self, filename, variables={}):
        """
        Processes the file stored at ``filename``. Any variables required by
        this template should be passed in through the dict ``variables``.
        """
        # Reset error log
        self.error = None

        # Check input
        if not type(variables) == dict:
            raise ValueError(
                'Second argument passed to process() must be dict'
                ' (variable_name : value)')

        # Convert script, any exceptions are thrown as PypeErrors
        script = self._convert(filename)

        # Get or create output stream
        stdout = self.stream if self.stream else io.StringIO()
        stderr = io.StringIO()

        # Run and handle errors
        error = None
        try:
            # Run, but catch any output + any occurring exception
            sysout = syserr = None
            try:
                sysout = sys.stdout
                syserr = sys.stderr
                sys.stdout = stdout
                sys.stderr = stderr
                exec(script, variables)
            except Exception:
                error = sys.exc_info()
            finally:
                if sysout:
                    sys.stdout = sysout
                if syserr:
                    sys.stderr = syserr    # Ignore error output

            if error:
                # Add exception traceback to error message
                msg = traceback.format_exception(*error)
                line = None

                # Search for error occurring in <string> (IE the executable
                # template)
                try:
                    # Try checking the last exception, see if its filename
                    # property is set and if it is equal to <string>
                    if error[1].filename == '<string>':
                        line = error[1].lineno  # pragma: no cover
                    else:
                        raise AttributeError
                except AttributeError:
                    # No filename or lineno property or filename not equal to
                    # <string>. Attempt to scan the strack trace for a frame
                    # with filename <string> and extract the line number.
                    next = error[2]
                    try:
                        while next:
                            if next.tb_frame.f_code.co_filename == '<string>':
                                line = next.tb_lineno
                                break
                            next = next.tb_next
                    finally:
                        del next

                if line:
                    # Error during template execution
                    sep = '- ' * 39 + '-'
                    msg.append(sep)
                    msg.append('An error occurred during execution of the'
                               ' processed template shown below.')
                    msg.append(sep)
                    lines = script.splitlines()
                    f = '{:>' + str(1 + len(str(line))) + '} '
                    for k, line in enumerate(lines[0:line]):
                        msg.append(f.format(1 + k) + line)
                    msg.append(sep)
                    msg.append(msg[0])
                self.error = '\n'.join(msg)
                print(self.error)
                raise PypeError('Error during template execution step.')

        finally:
            # Retrieving stack in except clause creates self reference in
            # stack, causing the garbage handler never to delete it. Solve
            # by deleting reference.
            # See: http://docs.python.org/library/sys.html#sys.exc_info
            if error:
                del error

        # Custom stream? Then don't interfere. If not, return stream contents.
        return None if self.stream else stdout.getvalue()

    def _convert(self, source):
        """
        Translates a pype file to a python file
        """
        with open(source, 'r') as f:
            source = f.read()

        # Define token recognising regex, helpers
        tags = [r'<\?=', r'<\?', r'\?>']
        rTags = re.compile(r'(' + '|'.join(tags) + ')')
        rQuot = re.compile(r'(""")')
        rEol = re.compile('[\n]{1}')
        rWhite = re.compile(r'[ \f\t]*')
        indent = ''

        # Convert
        tag_open = None
        out = ['import sys']
        for part in rTags.split(source):

            if part == '?>':
                if tag_open is None:
                    self.error = 'Closing tag found without opening tag'
                    raise PypeError(self.error)
                tag_open = None

            elif part == '<?' or part == '<?=':
                if tag_open is not None:
                    self.error = 'Nested opening tag found'
                    raise PypeError(self.error)
                tag_open = part

            elif tag_open == '<?':
                # Full python statement, remember final indenting
                line = rEol.split(part).pop()
                m = rWhite.match(line)
                indent = line[0:m.end()]
                out.append(part)

            elif tag_open == '<?=':
                # Quick printing statement, python code must be expression
                part = part.strip()
                try:
                    e = ast.parse(part)
                    assert len(e.body) == 1
                    assert isinstance(e.body[0], ast.Expr)
                except Exception:
                    msg = 'Code within <?= ?> tags can only contain a single' \
                          ' expression.'
                    err = traceback.format_exc().splitlines()
                    err.append(msg)
                    self.error = '\n'.join(err)
                    raise PypeError(msg)

                out.append(
                    indent + 'sys.stdout.write(str(' + part + '))')

            else:
                # Non-python code, just print
                # Triple quoted strings should be handled separately
                for part2 in rQuot.split(part):
                    if part2 == '"""':
                        out.append(indent + 'sys.stdout.write(\'"""\')')
                        continue

                    # If part ends in a ", this will cause problems, so...
                    nQuotes = 0
                    while part2[-1:] == '"':
                        part2 = part2[0:-1]
                        nQuotes += 1

                    out.append(
                        indent + 'sys.stdout.write(r"""' + part2 + '""")')

                    if nQuotes > 0:
                        out.append(
                            indent + 'sys.stdout.write(\'' + '"' * nQuotes
                            + '\')')

        return '\n'.join(out)

    def set_output_stream(self, stream):
        """
        When handling a template, all output will be directed into this stream.
        If no stream is specified, the standard output stream stdout is used.
        """
        self.stream = stream


class PypeError(Exception):
    """
    An error thrown by the :class:`TemplateEngine`

    *Extends:* Exception
    """
    def __init__(self, message):
        super().__init__(message)




#
# Factory methods for pacing protocols
#
# This file is part of Myokit.
# See http://myokit.org for copyright, sharing, and licensing details.
#
import myokit
"""
This module contains factory methods to create Protocol objects
programmatically.
"""


def blocktrain(period, duration, offset=0, level=1.0, limit=0):
    """
    Creates a train of block pulses.

    Each pulse lasts ``duration`` time units and a pulse is initiated every
    ``period`` time units.

    An optional offset to the first pulse can be specified using ``offset`` and
    the level of each pulse can be set using ``level``. To limit the number of
    pulses generated set limit to a non-zero value.
    """
    p = myokit.Protocol()
    #          level, start, duration, period=0, multiplier=0
    p.schedule(level, offset, duration, period, limit)
    return p


def bpm2bcl(bpm, m=1e-3):
    """
    Converts a beats-per-minute number to a basic cycle length in ms. For
    example a bpm of 60 equals a bcl of 1000ms.

    >>> import myokit
    >>> print(myokit.pacing.bpm2bcl(60))
    1000.0

    To use a different unit scaling, change the optional parameter ``m``. For
    example, with m set to 1 the returned units are seconds:

    >>> print(myokit.pacing.bpm2bcl(120, 1))
    0.5
    """
    return 60.0 / m / bpm


def constant(level):
    """
    Creates a very simple protocol where the pacing variable is held constant
    at a given level specified by the argument ``level``.
    """
    t = 1e9
    p = myokit.Protocol()
    p.schedule(level, 0, t, t, 0)
    return p


def steptrain(vsteps, vhold, tpre, tstep, tpost=0):
    """
    Creates a series of increasing or decreasing steps away from the fixed
    holding potential ``vhold``, towards the voltages listed in ``vsteps``.

      1. For the first ``tpre`` time units, the pacing variable is held at the
         value given by ``vhold``.
      2. For the next ``tstep`` time units, the pacing variable is held at a
         value from ``vsteps``
      3. For the next ``tpost`` time units, the pacing variable is held at
         ``vhold`` again.

    These three steps are repeated for each value in the ``vsteps``.
    """
    # Check time arguments
    if tpre < 0:
        raise ValueError('Time tpre can not be negative.')
    if tstep < 0:
        raise ValueError('Time tstep can not be negative.')
    if tpost < 0:
        raise ValueError('Time tpost can not be negative.')
    # Create protocol
    p = myokit.Protocol()
    time = 0
    for vstep in vsteps:
        if tpre > 0:
            p.schedule(vhold, time, tpre)
            time += tpre
        if tstep > 0:
            p.schedule(vstep, time, tstep)
            time += tstep
        if tpost > 0:
            p.schedule(vhold, time, tpost)
            time += tpost
    return p


def steptrain_linear(vstart, vend, dv, vhold, tpre, tstep, tpost=0):
    """
    Creates a series of increasing or decreasing steps away from a holding
    value (typically a holding potential). This type of protocol is commonly
    used to measure activation or inactivation in ion channel models.

      1. For the first ``tpre`` time units, the pacing variable is held at the
         value given by ``vhold``.
      2. For the next ``tstep`` time units, the pacing variable is held at a
         value ranging from vstart to vend, with increments dv.
      3. For the next ``tpost`` time units, the pacing variable is held at the
         value ``vhold`` again.

    These three steps are repeated for each value in the range from ``vstart``
    up to (but not including) ``vend``, with an increment specified as ``dv``.
    """
    # Check v arguments
    if vend > vstart:
        if dv <= 0:
            raise ValueError('vend > vstart so dv must be strictly positive.')
    else:
        if dv >= 0:
            raise ValueError('vend < vstart so dv must be negative.')
    # Check time arguments
    if tpre < 0:
        raise ValueError('Time tpre can not be negative.')
    if tstep < 0:
        raise ValueError('Time tstep can not be negative.')
    if tpost < 0:
        raise ValueError('Time tpost can not be negative.')
    # Create protocol
    p = myokit.Protocol()
    time = 0
    for i in range(int(abs((vend - vstart) / dv))):
        if tpre > 0:
            p.schedule(vhold, time, tpre)
            time += tpre
        if tstep > 0:
            p.schedule(vstart + i * dv, time, tstep)
            time += tstep
        if tpost > 0:
            p.schedule(vhold, time, tpost)
            time += tpost
    return p

